# Tektronix PWS4205 class, to perform the communication between the Wrapper and the device
# Sebastian Probst, sebastian.probst@kit.edu 2013
#
# based on the work on the Keithley_2636A driver by
# Pieter de Groot <pieterdegroot@gmail.com>, 2008
# Martijn Schaafsma <qtlab@mcschaafsma.nl>, 2008
# Hannes Rotzinger, hannes.rotzinger@kit.edu 2010
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import time
import logging
import numpy as np

class Tektronix_PWS4205(Instrument):
    '''
    This is the driver for the Tektronix_PWS4205 Power Supply

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'Tektronix_PWS4205', address='<USB address>, reset=<bool>')
    USB address e.g. USB0::0x0699::0x0390::C010427::INSTR
    '''

    def __init__(self, name, address, reset=False):
        '''
        Initializes the Tektronix_PWS4205, and communicates with the wrapper.

        Input:
          name (string)    : name of the instrument
          address (string) : GPIB or USB address
          reset (bool)     : resets to default values, default=False
        '''
        logging.info(__name__ + ' : Initializing instrument Tektronix_PWS4205')
        Instrument.__init__(self, name, tags=['physical'])

        # Add some global constants
        self._address = address
        self._visainstrument = visa.instrument(self._address)
        self._name = name
        
        self.add_parameter('voltage', flags=Instrument.FLAG_GETSET, units='V', minval=0, maxval=20, type=float)
        self.add_parameter('setvoltage', flags=Instrument.FLAG_GET, units='V', minval=0, maxval=20, type=float)
        self.add_parameter('current', flags=Instrument.FLAG_GETSET, units='A', minval=0, maxval=5, type=float)
        self.add_parameter('setcurrent', flags=Instrument.FLAG_GET, units='A', minval=0, maxval=5, type=float)
        self.add_parameter('status', flags=Instrument.FLAG_GETSET, type=bool)

        self.add_function('reset')
        self.add_function('on')
        self.add_function('off')
        self.add_function('ramp_current')
 
#        self.add_function('exec_tsl_script')
#        self.add_function('exec_tsl_script_with_return')
#        self.add_function('get_tsl_script_return')
        
        #self.add_function ('get_all')

        #self._visainstrument.write('beeper.beep(0.5,1000)')
        if (reset):
            self.reset()
        else:
            self.get_all()

    def reset(self):
        '''
        Resets the instrument to default values

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : resetting instrument')
        self._visainstrument.write("*RST")
        self.get_all()

    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : get all')
        self.get('voltage')
        self.get('current')
        self.get('status')
        self.get('setcurrent')
        self.get('setvoltage')
    
#    def exec_tsl_script(self,script):
#        '''
#        Writes the script to the device
#        Input:
#            TSL script (basically lua)
#        Output:
#            None
#        '''
#        logging.debug(__name__ + ' : execute TSL script on device')
#        self._visainstrument.write(script)
        
#    def exec_tsl_script_with_return(self,script):
#        '''
#        Writes the script to the device
#        Input:
#            TSL script (basically lua)
#        Output:
#            None
#        '''
#        logging.debug(__name__ + ' : execute TSL script on device and return data')
#        return self._visainstrument.ask(script).strip()

 
#    def get_tsl_script_return(self):
#        '''
#        Writes the script to the device
#        Input:
#            None
#        Output:
#            data, can be anything
#        '''
#        logging.debug(__name__ + ' : return data from executed TSL script')
#        return self._visainstrument.read().strip()

 
    def do_get_voltage(self):
        '''
        Reads the voltage signal from the instrument

        Input:
            None

        Output:
            volt (float) : Voltage in Volts
        '''
        logging.debug(__name__ + ' : get voltage')
        return float(self._visainstrument.ask('MEAS:VOLT?'))
        
    def do_get_setvoltage(self):
        '''
        Reads the setvoltage signal from the instrument

        Input:
            None

        Output:
            volt (float) : Voltage in Volts
        '''
        logging.debug(__name__ + ' : get setvoltage')
        return float(self._visainstrument.ask('VOLT?'))

    def do_set_voltage(self, volt):
        '''
        Sets the Amplitude of the signal

        Input:
            volt (float) : voltage in volt

        Output:
            None
        '''
        logging.debug(__name__ + ' : set voltage to %f' % volt)
        self._visainstrument.write('VOLT %s' % volt)

    def do_get_current(self):
        '''
        Reads the current signal from the instrument

        Input:
            None

        Output:
            current (float) : current in amps
        '''
        logging.debug(__name__ + ' : get current')
        return float(self._visainstrument.ask('MEAS:CURR?'))
        
    def do_get_setcurrent(self):
        '''
        Reads the setcurrent signal from the instrument

        Input:
            None

        Output:
            current (float) : current in amps
        '''
        logging.debug(__name__ + ' : get setcurrent')
        return float(self._visainstrument.ask('CURR?'))

    def do_set_current(self, curr):
        '''
        Sets the Amplitude of the signal

        Input:
            curr (float) : current in amps

        Output:
            None
        '''
        logging.debug(__name__ + ' : set current to %f' % curr)
        self._visainstrument.write('CURR %s' % curr)

    def do_get_status(self):
        '''
        Reads the output status from the instrument

        Input:
            None

        Output:
            status (boolean) 
        '''
        logging.debug(__name__ + ' : get status')
        return bool(int(self._visainstrument.ask('OUTPUT?')))


    def do_set_status(self, status):
        '''
        Sets the output status of the instrument

        Input:
            status (boolean)

        Output:
            None
        '''
        logging.debug(__name__ + ' : set status to %d' % status)
        self._visainstrument.write('OUTPUT %d' %status)


    # shortcuts
    def off(self):
        '''
        Sets status to False

        Input:
            None

        Output:
            None
        '''
        self.set_status(False)

    def on(self):
        '''
        Sets status to True

        Input:
            None

        Output:
            None
        '''
        self.set_status(True)

    def ramp_current(self, target, step, wait=0.2, showvalue=True):
        '''
        Ramps the Amplitude of the signal

        Input:
            target (float) : target current in amps
            step (float) : current steps in amps
            wait (floa) : waiting time in sec
            showvalue (bool) : print(values)

        Output:
            None
        '''
        start = self.get_current()
        if(target < start): step = -step
        a = np.concatenate( (np.arange(start, target, step)[1:], [target]) )
        for i in a:
            volt = self.get_voltage()
            if showvalue: 
                print(self._name+":  I="+str(i) + "A  " + "V=" + str(volt) + "V")
            self.set_current(i)
            time.sleep(wait)



# Tabor_WX1284C.py driver, to perform the communication between the Wrapper and the device
# Lukas Gruenhaupt <gruenhaupt@gmail.com>, 04/2015
# Andre Schneider <andre.schneider@kit.edu>, 01/2016
#
# based on Tektronix_AWG7062.py class coded by
# Pieter de Groot <pieterdegroot@gmail.com>, 2008
# Martijn Schaafsma <qtlab@mcschaafsma.nl>, 2008
# Guenevere Prawiroatmodjo <guen@vvtp.tudelft.nl>, 2009
# Markus Jerger <jerger@kit.edu>, 2010
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
import numpy
import struct
import time

import pyvisa.constants as vc


class Tabor_WX1284C(Instrument):
    '''
    This is the python driver for the Tabor Electronics Arbitrary Waveform Generator WX1284C
    '''

    def __init__(self, name, address, chpair=None, reset=False, clock=6e9, numchannels = 4):
        '''
        Initializes the AWG WX1284C
        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
            reset (bool)     : resets to default values, default=false
            chpair (int)     : number of channel pair (1 for channels 1&2, 2 for channels 3&4)
        Output:
            None
        '''
        logging.debug(__name__ + ' : Initializing instrument at '+address)
        Instrument.__init__(self, name, tags=['physical'])




        self._address = address
        if visa.qkit_visa_version == 1:
            self._visainstrument = visa.instrument(self._address,term_chars = "\r\n")
            self._visainstrument.timeout=2
        else:
            self._visainstrument = visa.instrument(self._address)
            self._visainstrument.timeout=2000L
            self._visainstrument.visalib.set_buffer(self._visainstrument.session, vc.VI_READ_BUF, 4000)
            self._visainstrument.visalib.set_buffer(self._visainstrument.session, vc.VI_WRITE_BUF, 32000)

            self._visainstrument.read_termination = '\n'
            self._visainstrument.write_termination = '\n'

                #intf_type = self._visainstrument.get_visa_attribute(vc.VI_ATTR_INTF_TYPE)

            if self._visainstrument.get_visa_attribute(vc.VI_ATTR_INTF_TYPE) in (vc.VI_INTF_USB, vc.VI_INTF_GPIB, vc.VI_INTF_TCPIP):
                self._visainstrument.set_visa_attribute(vc.VI_ATTR_WR_BUF_OPER_MODE, vc.VI_FLUSH_ON_ACCESS)
                self._visainstrument.set_visa_attribute(vc.VI_ATTR_RD_BUF_OPER_MODE, vc.VI_FLUSH_ON_ACCESS)
                if self._visainstrument.get_visa_attribute(vc.VI_ATTR_INTF_TYPE) == vc.VI_INTF_TCPIP:
                    self._visainstrument.set_visa_attribute(vc.VI_ATTR_TERMCHAR_EN, vc.VI_TRUE)

            self._visainstrument.clear()


        
        
        
        self._ins_VISA_INSTR = False#(self._visainstrument.resource_class == "INSTR") #check whether we have INSTR or SOCKET type
        
        self._values = {}
        self._values['files'] = {}
        self._clock = clock

        if chpair not in [None, 1, 2]:
            raise ValueError("chpair should be 1, 2 or None.")

        if chpair and numchannels!=2:
            raise ValueError("numchannels should be 2.")

        if chpair is None and numchannels !=4:
            raise ValueError("numchannels should be 4.")

        self._numchannels = numchannels
        self._choff = 0

        if chpair == 2:
            self._choff = 2

        if numchannels == 4:
            self.add_parameter('runmode', type=str,
                flags=Instrument.FLAG_GETSET,
                channels=(1, 2), channel_prefix='p%d_')
        else:
            self.add_parameter('runmode', type=str,
                flags=Instrument.FLAG_GETSET)
        
        self.add_parameter('output', type=bool,
            flags=Instrument.FLAG_GETSET,
            channels=(1, self._numchannels), channel_prefix='ch%d_')
        self.add_parameter('marker_output', type=bool,
            flags=Instrument.FLAG_GETSET,
            channels=(1, self._numchannels), channel_prefix='ch%d_')
        self.add_parameter('trigger_impedance', type=float,       #changed, checked in manual
            flags=Instrument.FLAG_GETSET,
            minval=50, maxval=10e3, units='Ohm')
        self.add_parameter('trigger_level', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=-5, maxval=5, units='Volts')
        self.add_parameter('clock', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=1e6, maxval=1.25e9, units='Hz')
        self.add_parameter('reference_source', type=str,
            flags=Instrument.FLAG_GETSET)
        self.add_parameter('reference_source_freq', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=10e6, maxval=100e6, units='Hz')
        self.add_parameter('common_clock', type=bool,
                flags=Instrument.FLAG_GETSET)
        self.add_parameter('amplitude', type=float,
            flags=Instrument.FLAG_GETSET,
            channels=(1, self._numchannels), minval=0, maxval=2, units='Volts', channel_prefix='ch%d_')
        self.add_parameter('offset', type=float,
            flags=Instrument.FLAG_GETSET,
            channels=(1, self._numchannels), minval=-2, maxval=2, units='Volts', channel_prefix='ch%d_')
        self.add_parameter('marker_high', type=float,
            flags=Instrument.FLAG_GETSET,
            channels=(1, self._numchannels), minval=-2, maxval=2, units='Volts', channel_prefix='ch%d_')
        self.add_parameter('status', type=bool,
            flags=Instrument.FLAG_GETSET,
            channels=(1, self._numchannels),channel_prefix='ch%d_')
        self.add_parameter('sync_position', type=float,
                flags=Instrument.FLAG_GETSET,units='s')

        if numchannels == 4:
            self.add_parameter('trigger_mode', type=str,
                flags=Instrument.FLAG_GETSET,
                channels=(1, 2), channel_prefix='p%d_')
            self.add_parameter('trigger_delay', type=float,
                flags=Instrument.FLAG_GETSET,
                channels=(1, 2), channel_prefix='p%d_')
            self.add_parameter('sequence_mode', type=str,
                flags=Instrument.FLAG_GETSET,
                channels=(1, 2), channel_prefix='p%d_')
            self.add_parameter('trigger_source', type=str,
                flags=Instrument.FLAG_GETSET,
                channels=(1, 2), channel_prefix='p%d_')
            self.add_parameter('trigger_time', type=float,
                flags=Instrument.FLAG_GETSET,
                channels=(1, 2), channel_prefix='p%d_')
            self.add_parameter('sync_output', type=bool,
                flags=Instrument.FLAG_GETSET,
                channels=(1, 2), channel_prefix='p%d_')
            self.add_parameter('trigger_slope', type=str,
                flags=Instrument.FLAG_GETSET,
                channels=(1, 2), channel_prefix='p%d_')

        else:
            self.add_parameter('trigger_mode', type=str,
                flags=Instrument.FLAG_GETSET)
            self.add_parameter('trigger_delay', type=float,
                flags=Instrument.FLAG_GETSET)
            self.add_parameter('sequence_mode', type=str,
                flags=Instrument.FLAG_GETSET)
            self.add_parameter('trigger_source', type=str,
                flags=Instrument.FLAG_GETSET)
            self.add_parameter('trigger_time', type=float,
                flags=Instrument.FLAG_GETSET)
            self.add_parameter('sync_output', type=bool,
                flags=Instrument.FLAG_GETSET)
            self.add_parameter('trigger_slope', type=str,
                flags=Instrument.FLAG_GETSET)

        # Add functions
        self.add_function('reset')
        self.add_function('fix')
        self.add_function('wait')
        self.add_function('get_all')
        self.add_function('set_seq_length')
        self.add_function('clear_waveforms')

        self.add_function('ask') #for debug
        self.add_function('write') #for debug
        self.add_function('define_sequence')
        self.add_function('close')


        if numchannels==2:
            self.add_function('preset_readout')
            self.add_function('preset_manipulation')

        self.add_function('wfm_send')
        self.add_function('wfm_send2')
        
        if reset:
            self.reset()
        else:
            self.get_all()    
        self.write(":INST1;:MARK:SOUR USER;:INST3;:MARK:SOUR USER")

        print("The device is set up with %i channels. You use the channels %s" %(numchannels, str(range(self._choff+1, numchannels + self._choff+1))))

    # Functions
    def fix(self,verbose=True):
        '''
        Communication with the device is sometimes problematic. This functions helps by clearing the queue and reading/resetting the error memory.
        '''
        if not self._ins_VISA_INSTR:
            tmo= self._visainstrument.timeout
            self._visainstrument.timeout = .1
            if verbose: print "Emptying Queue"
            try:
                for i in range(200):
                    rl = str(self._visainstrument.read())
                    if verbose: print "In Buffer %i: "%i+rl
            except visa.VisaIOError:
                if verbose: print "Timeout after %i iterations"%i
            self._visainstrument.timeout = tmo
        if verbose: print "Clearing Error Memory:"
        for i in range(200):
            try:
                if self.check():
                    return i
            except ValueError as e:
                logging.warning(e)
                continue

    def check(self):
        #this function is used to check whether the command was successfull
        try:
            rv = self.ask(":SYST:ERR?")
            if rv[0]=="0": return True
            else: raise ValueError(__name__ + "(" +self.get_name()+") : Device responded with error ->%s<-"%rv)
        except visa.VisaIOError:
            return False

    def reset(self):
        '''
        Resets the instrument to default values
        Input:
            None
        Output:
            None
        '''
        logging.info(__name__ + ' : Resetting instrument')
        self.write('*RST')
        self.write(":INST1;:MARK:SOUR USER;:INST3;:MARK:SOUR USER")

    def preset(self):
        '''
        resets the instrument and then sets all parameters to the values necessary for the time domain setup
        '''
        self.reset()
        self.set_trigger_impedance(50)
        self.set_trigger_level(.2)
        #self.set_p1_trigger_mode('TRIG')
        #self.set_p2_trigger_mode('TRIG')
        #self.set_p1_runmode('SEQ')
        #self.set_p2_runmode('USER')
        #self.set_p1_sequence_mode('STEP')
        #self.set_p2_sequence_mode('STEP')
        for i in range(1,self._numchannels + 1):
            self.set('ch%i_amplitude'%i,2)
            self.set('ch%i_marker_output'%i,True)
            self.set('ch%i_marker_high'%i,1)
        self.get_all()

    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.
        Input:
            None
        Output:
            None
        '''
        self.fix(False)
        logging.info(__name__ + ' : Reading all data from instrument')
        #self.check()
        try:
            pass
            #self.get_trigger_impedance()
        except Exception:
            logging.warning('command trigger impedance not supported')
        self.get_trigger_level()
        self.get_clock()
        if self._numchannels == 4:
            for i in (1,2):
                self.get('p%d_runmode' % i)
                self.get('p%d_trigger_mode' % i)
                self.get('p%d_sequence_mode' % i)
                self.get('p%d_sync_output' %i)
                self.get('p%d_trigger_time' %i)
                self.get('p%d_trigger_delay' %i)
                self.get('p%d_trigger_source' %i)
                self.get('p%d_trigger_slope' %i)
        for i in range(1,self._numchannels + 1):
            self.get('ch%d_amplitude' % i)
            self.get('ch%d_offset' % i)
            self.get('ch%d_marker_high' % i)
            self.get('ch%d_marker_output' % i)
            self.get('ch%d_status' % i)
            self.get('ch%d_output' % i)
        if self._numchannels == 2:
            self.get('runmode')
            self.get('trigger_mode')
            self.get('sequence_mode')
            self.get('sync_output')
            self.get('trigger_time')
            self.get('trigger_delay')
            self.get('trigger_source')
            self.get('trigger_slope')

    def preset_readout(self, tr_slope = 'NEG'):
        '''
        prepare the specified AWG for readout mode

        inputs:
            trigger edge
        output:
            None
        '''
        self.set_runmode('USER')
        self.set_ch1_output(True)
        self.set_ch2_output(True)
        self.set_trigger_mode('TRIG')
        self.set_trigger_source('EXT')
        for i in range(1,self._numchannels + 1):
            self.set('ch%i_amplitude'%i,2)
            self.set('ch%i_marker_output'%i,True)
            self.set('ch%i_marker_high'%i,1)
        try:
            self.set_trigger_slope(tr_slope)
        except Exception as m:
            print 'Trigger slope has to be POS or NEG. Setting to NEG.', m
            self.set_trigger_slope('NEG')

    def preset_manipulation(self):
        '''
        prepare the specified AWG for manipulation
        default repitition time is 2e-4
        Input:
            None
        Output:
            None
        '''
        self.set_runmode('SEQ')
        self.set_ch1_output(True)
        self.set_ch2_output(True)
        self.set_trigger_mode('TRIG')
        self.set_trigger_source('TIM')
        self.set_sync_output(True)
        self.set_trigger_time(2e-4)
        self.set_sequence_mode("STEP")
        for i in range(1,self._numchannels + 1):
            self.set('ch%i_amplitude'%i,2)
            self.set('ch%i_marker_output'%i,True)
            self.set('ch%i_marker_high'%i,1)


    def ask(self,cmd):
        for i in range(15):
            try:
                logging.debug(__name__ + "(" +self.get_name()+" ask command:"+cmd[:100]+":")
                return self._visainstrument.ask(cmd).strip()
            except visa.VisaIOError as e:
                logging.debug(__name__ + "(" +self.get_name()+"): VisaIOError, retry ->%s<-"%e)
        raise ValueError(__name__ + "(" +self.get_name()+"): ask('%s') not successful after 15 retries ->%s<-"%(cmd[0:100],e))

    def write(self,cmd):
        for i in range(15):
            try:
                logging.debug(__name__ + "(" +self.get_name()+" write command:"+cmd[:100]+":")
                return self._visainstrument.write(cmd)
            except visa.VisaIOError as e:
                logging.debug(__name__ + "(" +self.get_name()+") :VisaIOError, retry ->%s<-"%e)
        raise ValueError(__name__ + "(" +self.get_name()+") : write('%s') not successful after 15 retries ->%s<-"%(cmd[0:100],e))

    def write_raw(self,cmd):
        if visa.qkit_visa_version == 1:
            self.write(cmd)
        else:
            self._visainstrument.write_raw(cmd)


    def clear_waveforms(self):
        '''
        Clears the waveform on all channels.
        Input:
            None
        Output:
            None
        '''
        logging.debug(__name__ + ' : Clear waveforms from channels')
        for idx in range(self._choff+1,self._numchannels + 1):
            self.write(':INST%i; :TRAC:DEL:ALL'%idx)

    def run(self):
        '''
        Initiates the output of a waveform or a sequence. This is equivalent to pressing
        Run/Delete/Stop button on the front panel. The instrument can be put in the run
        state only when output waveforms are assigned to channels.
        Input:
            None
        Output:
            None
        '''
        logging.debug(__name__ + ' : Run/Initiate output of a waveform or sequence')
        self.write(':ENAB')

    def stop(self):
        '''
        Terminates the output of a waveform or a sequence. This is equivalent to pressing
        Run/Delete/Stop button on the front panel.
        Input:
            None
        Output:
            None
        '''
        logging.debug(__name__ + ' : Stop/Terminate output of a waveform or sequence')
        self.write('ABOR')

    def wait(self, max_timeouts = 1, quiet = False):
        '''
        Wait until the previous command has completed.

        Input:
            timeouts - maximum number of timeouts that may occur. note that the command
                does not execute at all if zero maximum timeouts are requested.
            quiet - if True, do not show messages when a timeout occurs
        '''
        if(max_timeouts > 0):
            self.write('*STB?')
        for i in range(max_timeouts):
            try:
                time.sleep(0.025)
                self._visainstrument.read()
                return
            except:
                if(not quiet): print('still waiting for the awg to become ready')
                time.sleep(1)

    def do_set_output(self, state, channel):
        '''
        This command sets the output state of the AWG.
        Input:
            channel (int) : the source channel
            state (bool) : True or False
        Output:
            None
        '''
        return self.set("ch%i_status"%channel,state)

    def do_get_output(self, channel):
        '''
        This command gets the output state of the AWG.
        Input:
            channel (int) : the source channel
        Output:
            state (bool) : True or False
        '''
        return self.get('ch%d_status'%channel)

    def do_set_marker_output(self,state,channel):
        channel +=self._choff
        logging.debug(__name__ + ' : Set marker %i output state %s'%(channel,state))
        self.write(':INST%s;:MARK:SEL%i;:MARK:STAT %i' % (channel,((int(channel)-1)%2+1),state))
        self.check()

    def do_get_marker_output(self,channel):
        channel +=self._choff
        logging.debug(__name__ + ' : Get marker %i output state '%(channel))
        outp = self.ask(':INST%s;:MARK:SEL%i;:MARK:STAT?' % (channel,((int(channel)-1)%2+1)))
        if ((outp=='0')|(outp == 'OFF')):
            return False
        elif ((outp=='1')|(outp == 'ON')):
            return True
        else:
            logging.debug(__name__ + ' : Read invalid status from instrument %s' % outp)
            raise ValueError('an error occurred while reading status from instrument')

    def do_set_runmode(self,mode,channel=1):
        '''
        Set the output mode of the AWG:
        FIX - Standard waveform shapes
        USER - Arbitrary waveforms
        SEQ - Sequenced mode
        ASEQ - Advanced seq. mode
        MOD - Modulated waveforms
        PULS
        PATT
        '''
        channel +=self._choff
        mode = mode.upper()
        if not mode in ("FIX","USER","SEQ","ASEQ","MOD","PULS","PATT"):
            raise ValueError("The selected mode is not supported, your choice was %s, but I have FIX,USER,SEQ,ASEQ,MOD,PULS,PATT"%mode)
        return self.write(":INST%s;:FUNC:MODE %s"%((2*channel-1 if self._numchannels == 4 else channel),mode))

    def do_get_runmode(self,channel=1):
        '''
        returns the output mode of the AWG.
        '''
        channel +=self._choff
        return self.ask(":INST%s;:FUNC:MODE?"%(2*channel-1 if self._numchannels == 4 else channel))


    def do_set_sequence_mode(self,mode,channel=1):
        '''
        Sequence Mode:
        AUTO - repeatedly run through the whole sequence (ignoring trigger)
        ONCE - play whole sequence once at trigger
        STEP - play one step per trigger event
        '''
        channel +=self._choff
        mode = mode.upper()
        if not mode in ("AUTO","ONCE","STEP"):
            raise ValueError("The selected sequence mode is not supported, your choice was %s, but I have AUTO,ONCE,STEP"%mode)
        return self.write(":INST%s;:SEQ:ADV %s"%((2*channel-1 if self._numchannels == 4 else channel),mode))

    def do_get_sequence_mode(self,channel=1):
        channel +=self._choff

        return self.ask(":INST%s;:SEQ:ADV?"%(2*channel-1 if self._numchannels == 4 else channel))



    def do_set_trigger_mode(self, runmode,channel=1):
        '''
        Set the Trigger Mode of the device to Continuous, Triggered or Gated.
        Input:
            runmode (str) : The Trigger mode which can be set to 'CONT', 'TRIG', 'GAT'.
        Output:
            None
        '''
        channel +=self._choff
        logging.debug(__name__ + ' : Set runmode to %s' % runmode)
        runmode = runmode.upper()
        if (runmode == 'TRIG'):
            self.write(':INST%s;:INIT:CONT 0;GATE 0'%(2*channel-1 if self._numchannels == 4 else channel))
        else:
            if (runmode == 'CONT'):
                self.write(':INST%s;:INIT:CONT 1'%(2*channel-1 if self._numchannels == 4 else channel))
            else:
                if (runmode == 'GATE'):
                    self.write(':INST%s;:INIT:CONT 0; GATE 1'%(2*channel-1 if self._numchannels == 4 else channel))
                else:
                    logging.error(__name__ + ' : Unable to set trigger mode to %s, expected "CONT", "TRIG" or "GATE"' % runmode)

    def do_get_trigger_mode(self,channel=1):
        '''
        Get the Trigger Mode of the device
        Output:
            runmode (str) : The Trigger mode which can be set to 'CONT', 'TRIG' or 'GAT'.
        '''
        channel +=self._choff
        cont = self.ask(':INST%s;:INIT:CONT ?'%(2*channel-1 if self._numchannels == 4 else channel))

        if (cont == 'OFF'):
            gate = self.ask(':INST%s;:INIT:GATE ?'%(2*channel-1 if self._numchannels == 4 else channel))
            if (gate == 'OFF'):
                return 'TRIG'
            else:
                return 'GATE'
        else:
            return 'CONT'

    def do_get_trigger_delay(self,channel=1):
        '''
        gets the trigger delay for the specified pair of channels. Internally, the delay is counted in samples, but this function returns seconds.
        '''
        channel +=self._choff
        return float(self.ask(":INST%s;:TRIG:DEL?"%(2*channel-1 if self._numchannels == 4 else channel)))/float(self.get_clock())

    def do_set_trigger_delay(self,delay,channel=1):
        '''
        gets the trigger delay for the specified pair of channels. Internally, the delay is counted in samples, but this function returns seconds.
        '''
        channel +=self._choff
        clock = self.get_clock()
        self.write(":INST%s;:TRIG:DEL%i"%((2*channel-1 if self._numchannels == 4 else channel),round(delay*clock)))

    def do_set_trigger_source(self, source, channel=1):
        '''
        sets the source of the trigger signal external or internal (timer)
        Input:
            EXT: external trigger input
            TIM: internal trigger
        Output:
            None
        '''
        channel +=self._choff
        source = source.upper()
        if source == "INT":
            source = "TIM"
            logging.warning("I assume you mean TIM. I fixed this for you :)")
        if not source in ("EXT","TIM"):
            raise ValueError("The selected source is not supported, your choice was %s, but I have EXT or TIM"%source)

        return self.write(":INST%s;:TRIG:SOUR:ADV %s"%((2*channel-1 if self._numchannels == 4 else channel),source))

    def do_get_trigger_source(self, channel=1):
        '''
        returns the source of the trigger signal external or internal (timer)
        Input:
            None
        Output:
            EXT: external trigger input
            TIM: internal trigger
        '''
        channel +=self._choff
        logging.debug(__name__ + ' : Get source of channel %s' % (2*channel-1 if self._numchannels == 4 else channel))
        return self.ask(":INST%s;:TRIG:SOUR:ADV?" %(2*channel-1 if self._numchannels == 4 else channel))

    def do_set_trigger_time(self, time, channel=1):
        '''
        sets the repitition time of the trigger signal
        Input:
            time
        Output:
            None
        '''
        channel +=self._choff

        return self.write(":INST%s;:TRIG:TIM:TIME %e"%((2*channel-1 if self._numchannels == 4 else channel),time))

    def do_get_trigger_time(self, channel = 1):
        '''
        gets the repitition time of the trigger signal
        Input:
            None
        Output:
            time
        '''
        channel +=self._choff
        logging.debug(__name__ + ' : Get time of channel %s' % (2*channel-1 if self._numchannels == 4 else channel))
        return self.ask(":INST%s;:TRIG:TIM:TIME?" % (2*channel-1 if self._numchannels == 4 else channel))

    def do_set_sync_output(self, state, channel = 1):
        '''
        sends a sync pulse to sync out, can be used for synchronisation with chpair 2
        Input:
            True or False
        Output:
            None
        '''
        channel+=self._choff
        self.write(":OUTP:SYNC:SOUR%s"%(2*channel-1 if self._numchannels == 4 else channel))
        self.write(":OUTP:SYNC%i"%state)


    def do_get_sync_output(self, channel=1):
        '''
        checks if there is a synchronisation pulse
        Input:
            None
        Output:
            On or Off
        '''
        channel +=self._choff
        if self._numchannels == 4:
            channel = 2*channel-1
        if int(self.ask(":OUTP:SYNC:SOUR ?"))==channel:
            if self.ask(":OUTP:SYNC ?")=="ON":
                return True
        return False
    
    def do_set_sync_position(self,position):
        '''
        Set the position of the SYNC output pulse.
        Input:
            position: start of the pulse after the start of the waveform in seconds. This is internally converted into samples and rounded to 32.
        '''
        samplepos = int(round(position * self.get_clock() /32))*32
        self.write(":OUTP:SYNC:POS%i"%samplepos)

    def do_get_sync_position(self):
        return float(self.ask(":OUTP:SYNC:POS?"))/self.get_clock()

    def set_trig_impedance(self, impedance):
        '''
        Sets the trigger impedance to 50 Ohm or 10 kOhm
        Input:
            '50' or '10k'
        Output:
            None
        '''
        logging.debug(__name__  + ' : Set trigger impedance')
        if ((impedance == '50') | (impedance == '10k')):
            self.write('TRIG:INP:IMP %s' % impedance)
        else:
            logging.error(__name__ + ' : Unable to set impedance to %s Ohm, expected "50" or "10k"' % impedance)


    def do_get_trigger_impedance(self):
        '''
        Reads the trigger impedance from the instrument
        Input:
            None
        Output:
            impedance: 10K or 50 (Ohm)
        '''
        #logging.error(__name__ + ' : getting trigger impedance does currently not work... (AS 01/2016)')
        #return 0
        logging.debug(__name__  + ' : Get trigger impedance from instrument')
        imp = self.ask('TRIG:INP:IMP ?')
        if imp == "10K": imp = 1e4
        return  imp

    def do_set_trigger_impedance(self, mod):
        '''
        Sets the trigger impedance of the instrument
        Input:
            mod (int) : Either 1e3 or 50 depending on the mode
        Output:
            None
        '''
        if (mod==10e3):
            self.set_trig_impedance('10k')
        elif (mod==50):
            self.set_trig_impedance('50')
        else:
            logging.error(__name__ + ' : Unable to set trigger impedance to %s, expected "10e3" or "50"' % mod)

    def do_get_trigger_level(self):
        '''
        Reads the trigger level from the instrument
        Input:
            None
        Output:
            None
        '''
        logging.debug(__name__  + ' : Get trigger level from instrument')
        return float(self.ask('TRIG:LEV ?'))

    def do_set_trigger_level(self, level):
        '''
        Sets the trigger level of the instrument
        Input:
            level (float) : trigger level in volts
        '''
        logging.debug(__name__  + ' : Trigger level set to %.2f' % level)
        self.write('TRIG:LEV %.2f' % level)

    def do_set_trigger_slope(self, slope, channel=1):
        '''
        Trigger starts with positive or negative slope of trigger signal
        Input:
            slope: 'POS', 'NEG' or 'EIT'
            channel
        Output:
            None
        '''
        slope=slope.upper()
        if slope in ["POS", "NEG", "EIT"]:
            channel+=self._choff
            self.write(":INST%i;:TRIG:SLOP%s"%((2*channel-1 if self._numchannels == 4 else channel),slope))

    def do_get_trigger_slope(self, channel=1):
        '''
        with which slope does the trigger start?
        Input:
            channel
        Output:
            slope
        '''
        return self.ask(":INST%i;:TRIG:SLOP?"%(2*channel-1 if self._numchannels == 4 else channel))

    def do_get_clock(self):
        '''
        Returns the clockfrequency, which is the rate at which the datapoints are
        sent to the designated output
        Input:
            None
        Output:
            clock (int) : frequency in Hz
        '''
        self._clock = self.ask(":FREQ:RAST?")
        return self._clock

    def do_set_clock(self, clock):
        '''
        Sets the rate at which the datapoints are sent to the designated output channel
        Input:
            clock (int) : frequency in Hz (75e6 to 1.25e9)
        Output:
            None
        '''
        self._clock = clock
        self.write(':FREQ:RAST%f' % clock)

    def do_get_amplitude(self, channel):
        '''
        Reads the amplitude of the designated channel from the instrument
        Input:
            channel (int) : 1, 2, 3 or 4: the number of the designated channel
        Output:
            amplitude (float) : the amplitude of the signal in Volts
        '''
        channel +=self._choff
        logging.debug(__name__ + ' : Get amplitude of channel %s from instrument'
            % channel)
        return float(self.ask(':INST%s;:VOLT ?' % channel))

    def do_set_amplitude(self, amp, channel):
        '''
        Sets the amplitude of the designated channel of the instrument
        Input:
            amp (float)   : amplitude in Volts (0.050 V to 2.000 V)
            channel (int) : 1, 2, 3 or 4: the number of the designated channel
        Output:
            None
        '''
        channel +=self._choff
        logging.debug(__name__ + ' : Set amplitude of channel %s to %.6f'
            % (channel, amp))
        if amp < 50e-3:
            amp = 50e-3
            print 'amplitude was set to 0.05 V, which is smallest possible voltage'
        if amp > 2:
            amp = 2.
            print 'amplitude was set to 2 V, which is highest possible voltage'

        self.write(':INST%s;:VOLT%.3f' % (channel, amp))

    def do_get_offset(self, channel):
        '''
        Reads the offset of the designated channel of the instrument
        Input:
            channel (int) : 1, 2, 3 or 4: the number of the designated channel
        Output:
            offset (float) : offset of designated channel in Volts
        '''
        channel +=self._choff
        logging.debug(__name__ + ' : Get offset of channel %s' % channel)
        return float(self.ask(':INST%s;:VOLT:OFFS ?' % channel))

    def do_set_offset(self, offset, channel):
        '''
        Sets the offset of the designated channel of the instrument
        Input:
            offset (float) : offset in Volts (-1.000 V to 1.000 V)
            channel (int)  : 1, 2, 3 or 4: the number of the designated channel
        Output:
            None
        '''

        channel +=self._choff
        if offset < -1.000:
            offset = -1.000
            print 'Offset was set to -1.000 V, which is smallest possible voltage'
        if offset > 1.000:
            offset = 1.000
            print 'Offset was set to 1.000 V, which is highest possible voltage'

        logging.debug(__name__ + ' : Set offset of channel %s to %.3f' % (channel, offset))
        self.write(':INST%s;:VOLT:OFFS%.3f' % (channel, offset))

    def do_get_marker_high(self, channel):
        '''
        Gets the high level for marker1 on the designated channel.
        Note that Channels 1&2 and 3&4 share the same two markers.
        Input:
            channel (int) : 1, 2, 3 or 4: the number of the designated channel
        Output:
            high (float) : high level in Volts
        '''
        channel +=self._choff
        logging.debug(__name__ + ' : Get upper bound of marker %i of channel %s' %(((int(channel)-1)%2+1),channel))
        return float(self.ask(':INST%s;:MARK:SEL%i;:MARK:VOLT:HIGH ?' % (channel,((int(channel)-1)%2+1))))

    def do_set_marker_high(self, high, channel):
        '''
        Sets the high level for marker1 on the designated channel.
        Note that Channels 1&2 and 3&4 share the same two markers.

        Input:
            high (float)   : high level in Volts
            channel (int)  : 1, 2, 3 or 4: the number of the designated channel
        Output:
            None
         '''
        channel +=self._choff
        logging.debug(__name__ + ' : Set upper bound of marker%i of channel %s to %.3f'% (((int(channel)-1)%2+1),channel, high))
        self.write(':INST%s;:MARK:SEL%i;:MARK:VOLT:HIGH%.2f' % (channel,((int(channel)-1)%2+1),high))
        self.check()

    def do_get_status(self, channel):
        '''
        Gets the status of the designated channel.
        Input:
            channel (int) : 1, 2, 3 or 4: the number of the designated channel
        Output:
            status (bool)
        '''
        channel +=self._choff
        logging.debug(__name__ + ' : Get status of channel %s' % channel)
        outp = self.ask(':INST%s;:OUTP ?' % channel)
        if ((outp=='0')|(outp == 'OFF')):
            return False
        elif ((outp=='1')|(outp == 'ON')):
            return True
        else:
            logging.debug(__name__ + ' : Read invalid status from instrument %s' % outp)
            return 'an error occurred while reading status from instrument'
        self.get("ch%i_output"%channel-self._choff)

    def do_set_status(self, status, channel):
        '''
        Sets the status of designated channel.
        Input:
            status (bool) : True or False
            channel (int)   : 1, 2, 3 or 4: the number of the designated channel
        Output:
            None
        '''
        channel +=self._choff
        logging.debug(__name__ + ' : Set status of channel %s to %s' % (channel, status))
        if status:
            self.write(':INST%s;:OUTP ON' % channel)
        else:
            self.write(':INST%s;:OUTP OFF' % channel)
        self.get("ch%i_output"%(channel-self._choff))

    def do_set_reference_source(self,source='EXT'):
        '''
        Sets the clock reference source.
        Input:
            source (string) : 'INT' or 'EXT'
        Output:
            None
        '''
        if source == 'EXT' or source == 'INT':
            logging.debug(__name__ + ' : Set clock reference source to %s' %source)
            self.write(':ROSC:SOUR %s' %source)
        else:
            logging.warning('Clock source needs to be one of INT or EXT. No changes made.')

    def do_get_reference_source(self):
        '''
        Gets the clock reference source.
        Input:
            None
        Output:
            (string) : 'INT' or 'EXT'
        '''
        logging.debug(__name__ + ' : Get clock reference source.')
        return str(self.ask(':ROSC:SOUR ?'))

    def do_set_reference_source_freq(self,freq=10e6):
        '''
        Sets the clock reference frequency.
        Input:
            freq (float) : 10e6, 20e6, 50e6, 100e6
        Output:
            None
        '''
        if freq in [10e6,20e6,50e6,100e6]:
            logging.debug(__name__ + ' : Set clock reference frequency to %.2g' %freq)
            self.write(':ROSC:FREQ %s' %str(freq))
        else:
            logging.warning('Clock source needs to be one of [10e6,20e6,50e6,100e6]. No changes made.')

    def do_get_reference_source_freq(self):
        '''
        Gets the clock reference frequency.
        Input:
            None
        Output:
            (string) : 10e6, 20e6, 50e6, 100e6
        '''
        logging.debug(__name__ + ' : Get clock reference frequency.')
        return float(self.ask(':ROSC:FREQ ?'))

    def do_set_common_clock(self,status=True):
        '''
        Use a common trigger clock for both channel pairs.
        Input:
            status (bool)
        Output:
            None
        '''
        if status:
            logging.debug(__name__ + ' : Set common clock.')
            self.write(':INST:COUPLE:STAT ON')
        else:
            logging.debug(__name__ + ' : Set clocks to be separate.')
            self.write(':INST:COUPLE:STAT OFF')

    def do_get_common_clock(self):
        '''
        Check whether a common clock for both channel pairs is used.
        Input:
            None
        Output:
            (bool)
        '''
        logging.debug(__name__ + ' : Check common clock.')
        return bool(self.ask(':INST:COUPLE:STAT ?')=='ON')

    def send_waveform(self, w, m1, m2, channel, seg):
        return self.wfm_send(w, m1, m2, channel, seg)

    def wfm_send(self, w, m1, m2, channel, seg):
        '''
        Sends a complete waveform. All parameters need to be specified.
        Takes a waveform as generated by generate_waveform from qkit and first converts it to data usable by the AWG
        then sends it to the memory of the specified channel into the specified segment.
        Minimum waveform size is 192 points. If a waveform is shorter, it will be appended automatically.

        See also: resend_waveform()
        Input:
            w (float[numpoints]) : waveform (amplitude range 0 <-max amp.> to 1 <max amp.>)
            m1 (int[numpoints])  : marker1 (either 1 or 0, i.e. marker on or off)
            m2 (int[numpoints])  : marker2 (either 1 or 0, i.e. marker on or off)
            use m1=None or m2=None to fill with zeros
            channel (int)        : channel 1, 2, 3, or 4 that waveform is sent to
            seg (int)            : # of data segment the waveform is sent to (1 to 16000)
        Output:
            None
        '''

        if channel > self._numchannels:
            raise ValueError("There are only channels 1 and 2.")

        channel +=self._choff

        dim = len(w)

        if len(w)%16 != 0:
            raise ValueError #wfm length has to be divisible by 16
        if len(w) < 192:
            w = numpy.append(w1, numpy.zeros(192 - len(w)))
            m1 = numpy.append(m1, numpy.zeros(192 - len(m1)))
            m2 = numpy.append(m2, numpy.zeros(192 - len(m2)))

        if(m1 == None): m1 = numpy.zeros_like(w)
        if(m2 == None): m2 = numpy.zeros_like(w)
        if (not((len(w)==len(m1)) and ((len(m1)==len(m2))))):
            raise ValueError("error, the length of your waveform and markes does not match")

        self.write(':TRAC:DEF%i,%i' % (seg,len(w)))

        #Set specified channel and number of memory segment
        self.write(':INST%s;:TRAC:SEL%i' % (channel, seg))

        #set set single trace transfer mode
        self.write(':TRAC:MODE SING')

        ws = ''
        for i in range(0,len(w)):
            ws = ws + struct.pack('<H', 8191*w[i]+8192+m1[i]*2**14+m2[i]*2**15)

        self.write_raw(':TRAC#%i%i%s' %(int(numpy.log10(len(ws))+1), len(ws), ws))

    def wfm_send2(self, w1, w2, m1=None, m2=None, channel=1, seg=1):
        '''
        Sends two complete waveforms for channel pairs 1&2 or 3&4. All parameters need to be specified.
        Takes a waveform as generated by generate_waveform from qkit and first converts it to data usable by the AWG
        then sends it to the memory of the specified channel into the specified segment.
        Minimum waveform size is 192 points. If the two waveforms are shorter, they will be appended automatically.

        See also: resend_waveform()
        Input:
            w (float[numpoints]) : waveform (amplitude range 0 <-max amp.> to 1 <max amp.>)
            m1 (int[numpoints])  : marker1 (either 1 or 0, i.e. marker on or off)
            m2 (int[numpoints])  : marker2 (either 1 or 0, i.e. marker on or off)
            channel (int)        : channels are paired (1&2, 3&4) selecting either of one pair is fine
            seg (int)            : # of data segment the waveform is sent to (1 to 16000)
        Output:
            None
        '''
        channel +=self._choff
        if len(w1) != len(w2): raise ValueError("Waveform length is not equal.")
        if len(w1)%16 != 0: raise ValueError("Wfm length has to be divisible by 16")
        if len(w1) < 192:
            w1 = numpy.append(w1, numpy.zeros(192 - len(w1)))
            w2 = numpy.append(w2, numpy.zeros(192 - len(w2)))
            m1 = numpy.append(m1, numpy.zeros(192 - len(m1)))
            m2 = numpy.append(m2, numpy.zeros(192 - len(m2)))

        if m1 is None : m1 = numpy.zeros_like(w1)
        if m2 is None : m2 = numpy.zeros_like(w1)
        if (not((len(w1)==len(m1)) and ((len(m1)==len(m2))))):
            raise ValueError("error, the length of your waveform and markes does not match")

        self.write(':TRAC:DEF%i,%i' % (seg,len(w1)))

        #Set specified channel and number of memory segment
        self.write(':INST%s;:TRAC:SEL%i' % (channel, seg))

        #set set combined trace transfer mode
        self.write(':TRAC:MODE COMB')

        wfm = numpy.append(numpy.reshape(numpy.array(8191*w2+8192+m1*2**14+m2*2**15,dtype=numpy.dtype('<H')),(-1,16)),numpy.reshape(numpy.array(8191*w1+8192+m1*2**14+m2*2**15,dtype=numpy.dtype('<H')),(-1,16)),axis=1).flatten()
        ws =str(buffer(wfm))

        self.write_raw(':TRAC#%i%i%s' %(int(numpy.log10(len(ws))+1), len(ws), ws))


    def define_sequence(self,channel, segments=None,loops=None,jump_flags=None):
        '''
        specifies the sequence table for sequenced mode.
        channel: segments table can be created for each pair of channels separately
        segments (array of ints or int): which segments in the awg do you want to output? (1-32000) If a simple int is specified, range(1,i+1) will be used
        loops (array of ints): How often should each segment be repeated? (1-16M)
        jump_flags (array of 0,1): if set to 1, the sequencer will wait at this sequence element until an event happens.
        '''
        channel +=self._choff
        if segments is None:
            print "Amount of segments not specified, try to get it from AWG"
            for i in range(1,32000):
                if int(self.ask(":TRAC:DEF%i?"%i).split()[-1])==0:
                    segments = i-1
                    break
            if segments is None:
                raise ValueError("Could not find number of segments...")
        if type(segments)==int:
            #the segement table needs at least 3 entries. So if it would be shorter, we just take it multiple times.
            if segments == 1 : segments = [1,1,1]
            elif segments == 2 : segments = [1,2,1,2]
            else: segments = range(1,segments+1)
        if loops == None: loops = numpy.ones(len(segments))
        if jump_flags == None: jump_flags = numpy.zeros(len(segments))
        if not len(loops)==len(segments) or not len(jump_flags) == len(segments):
            raise ValueError("Length of segments (%i) does not match length of loops (%i) or length of jump_flags(%i)"%(len(segments),len(loops),len(jump_flags)))
        if len(segments)<3: raise ValueError("Sorry, you need at least 3 segments. Your command has %i segments"%(len(segments)))

        #Set specified channel
        self.write(':INST%s' % (channel))
        ws = ''
        for i in range(len(segments)):
            ws = ws + struct.pack('<LHH', loops[i],segments[i],jump_flags[i])
        self.write_raw(':SEQ#%i%i%s' %(int(numpy.log10(len(ws))+1), len(ws), ws))

    def set_seq_length(self,length,chpair=1):
        self.define_sequence((2*chpair-1 if self._numchannels == 4 else chpair),length)

    def close(self):
        self._visainstrument.close()



# AS@KIT

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# install:
'''
    sudo apt-get update
    sudo apt-get install python-dev
    sudo apt-get install python-rpi.gpio
'''


import RPi.GPIO as GPIO
GPIO.setmode(GPIO.BOARD)   #use RPi.GPIO layout

#P = [19,   21, 23, 22, 24, 26] #rpi 2 #GPIO 10,9,11,25,8,7
P = [33,    35, 37, 36, 38, 40] #rpi 3 #GPIO 13,19,26,16,20,21
A = [0.5,   1,  2,  4,  8,  16]#Attenuation

GPIO.setup(P, GPIO.OUT, initial=GPIO.HIGH) #initialize to high to have maximum attenuation at start.

attenuation = sum(A)
A.reverse() #this is needed for the loop later
#P.reverse() #now they correspond to each other again

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer
from threading import Thread
from time import sleep


class TCPHandler_StepAttenuator(SocketServer.BaseRequestHandler):
    """
    The RequestHandler class for our server.

    It is instantiated once per connection to the server, and must
    override the handle() method to implement communication to the
    client.
    """

    def set_attenuation(self, att):
        global attenuation
        try:
            att=round(float(att),3) #the round is to avoid floating point precision issues. Would only affect udB settings.
        except Exception as m:
            print 'Conversion of attenuation to float failed: ', m
        
        set_values = [1]*len(A)
        att2 = att
        for i,a in enumerate(A):
            if att2 >= a:
                set_values[i]=0
                att2-=a
        if att2 != 0:
            print "Got request to set to %s dB, but I then would have %.4f dB left"%(att,att2)
            return False
        
        GPIO.output(P,set_values)
        attenuation = att
        print "Setting attenuation to %.3fdB: "%(att)+str(set_values)
        return True
            
    def handle(self):
        global attenuation
        try:
            # self.request is the TCP socket connected to the client
            self.data = str(self.request.recv(1024).strip()).upper()
            if "ATTEN" in self.data and "?" in self.data:
                print 'get attenuation request from', str(format(self.client_address[0]))
                self.request.sendall(str(attenuation))
            elif "ATTEN" in self.data:
                print 'Set attenuation request from', str(format(self.client_address[0]))
                att = float(filter(lambda x: str.isdigit(x) or x==".",self.data)) #extracts digits and decimal point
                self.request.sendall(str(self.set_attenuation(att)))
            else:
                print 'request string not recognized: %s'%self.data
                self.request.sendall(0)
        except Exception as m:
            print 'Error in TCP server handler:', m


class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):   #enable multiple access
    pass


if __name__ == "__main__":
    HOST, PORT = 'ip-address', 9900

    # Create the server, binding to localhost on port 9955
    server = ThreadedTCPServer((HOST, PORT), TCPHandler_StepAttenuator)

    # Activate the server; this will keep running until you
    # interrupt the program with Ctrl-C
    try:
        server.serve_forever()
    finally:
        print 'shut down server'
        server.shutdown()




# This file brings QKIT around: init 
# YS@KIT/2017
# HR@kit/2017
import qkit
import os
import importlib
import logging
from time import time

def start(silent=False):
    #print('Starting the core of the Qkit framework...')
    initdir_name = 's_init'
    initdir = os.path.join(qkit.cfg.get('coredir'),initdir_name)
    filelist = os.listdir(initdir)
    filelist.sort()
    
    # load all modules starting with a 'S' character
    for module in filelist:
        starttime = time()
        if not module.startswith('S') or module[-3:] != '.py':
            continue
        if not silent:
            print("Loading module ... "+module)
        importlib.import_module("."+module[:-3],package='qkit.core.'+initdir_name)
        logging.debug("Loading module "+str(module)+" took  {:.1f}s.".format(time()-starttime))
    del module,starttime
    


# -*- coding: utf-8 -*-
"""
@author: hannes.rotzinger@kit.edu / 2015,2016,2017, 2018
         marco.pfirrmann@kit.edu / 2016, 2017, 2018
@license: GPL

This file handles the actual display of the data within the respective canvas.
The functions here are called (case sensitive) by PlotWindow and the
information is mainly stored in self variables of the PlotWindow class object.
Possible display options are 1d (for data and views), 2d, table, as well as
JSON encoded and plain text.
The _display functions are supported by simple getter functions for metadata
readout.
"""
import sys

in_pyqt5 = False
try:
    from PyQt5.QtCore import Qt
    
    in_pyqt5 = True
except ImportError as e:
    pass

if not in_pyqt5:
    try:
        from PyQt4.QtCore import Qt
    except ImportError:
        print("import of PyQt5 and PyQt4 failed. Install one of those.")
        sys.exit(-1)

import numpy as np
import json
import pyqtgraph as pg
import qkit
from qkit.storage.hdf_constants import ds_types
import pprint
from qkit.core.lib.misc import str3


def _display_1D_view(self, graphicsView):
    """displays the 1d plot(s) views from the respective datasets.

    The entries in the overlay dict get evaluated and handled according to 
    their ds-type. Depending on the ds-type the QComboBox handles the index of 
    the shown data.

    Args:
        self: Object of the PlotWindow class.
        graphicsView: Object of pyqtgraph's PlotWidget class.

    Returns:
        No return variable. The function operates on an object of the 
        PlotWindow class.
    """
    
    graphicsView.clear()
    
    if not graphicsView.plotItem.legend:
        graphicsView.plotItem.addLegend(size=(160, 48), offset=(30, 15))
    
    overlay_num = self.ds.attrs.get("overlays", 0)
    
    ## evaluate the overlay-entries to get the ds_urls of the datasets to be 
    ## displayed.
    view_params = json.loads(self.ds.attrs.get("view_params", {}))
    for i in range(overlay_num + 1):
        xyurls = str3(self.ds.attrs.get("xy_" + str(i), ""))
        ds_urls = [xyurls.split(":")[0], xyurls.split(":")[1]]
        if xyurls:
            err_url = self.ds.attrs.get("xy_" + str(i) + "_error", "")
            if err_url:
                ds_urls.append(err_url)
            dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ds_urls)
            if dss[0] is None or dss[1] is None:
                print("Could not load view xy_%i: "%i + str(xyurls)+" : Datasets not found.")
                continue
            ## retrieve the data type and store it in  x_ds_type, y_ds_type
            x_ds_type = dss[0].attrs.get('ds_type', ds_types['coordinate'])
            y_ds_type = dss[1].attrs.get('ds_type', ds_types['coordinate'])
            
            ## If there are more than one dataset we need to check the dimensions
            ## first, the "lower" dimensional ds is the limiting factor.
            if x_ds_type == ds_types['coordinate'] or x_ds_type == ds_types['vector']:
                if y_ds_type == ds_types['vector'] or y_ds_type == ds_types['coordinate']:
                    self.VTraceXSelector.setEnabled(False)
                    self.VTraceYSelector.setEnabled(False)
                    
                    x_data = dss[0][()]
                    y_data = dss[1][()]
                    if err_url:
                        err_data = dss[2][()]
                    
                    # prevent a crash when the two datasets have a different length
                    # solution: truncate the longer dataset to the length of the shorter
                    # Note: This can happen if the hdf.flush() is not in sync with the viewer
                    
                    x_data_len = len(x_data)
                    y_data_len = len(y_data)
                    if x_data_len != y_data_len:
                        if x_data_len > y_data_len:
                            x_data = x_data[:y_data_len]
                        else:
                            y_data = y_data[:x_data_len]
                
                elif y_ds_type == ds_types['matrix']:
                    if view_params.get('transpose',False):
                        self.VTraceYSelector.setEnabled(True)
                        range_max = dss[1].shape[1]
                        if self.VTraceXSelector.isEnabled(): #Hack to see if the window is freshly generated
                            self.VTraceXSelector.setEnabled(False)
                            self.VTraceYSelector.setValue(view_params.get('default_trace',range_max //2))
                        self.VTraceYSelector.setRange(-1 * range_max, range_max - 1)
                        self.VTraceYValue.setText(self._getYValueFromTraceNum(dss[1], self.VTraceYNum))
                        x_data = dss[0][()]
                        y_data = dss[1][:,self.VTraceYNum]
                        if err_url:
                            err_data = dss[2][:,self.VTraceYNum]
                    else:
                        self.VTraceXSelector.setEnabled(True)
                        range_max = dss[1].shape[0]
                        self.VTraceXSelector.setRange(-1 * range_max, range_max - 1)
                        self.VTraceXValue.setText(self._getXValueFromTraceNum(dss[1], self.VTraceXNum))
                        self.VTraceYSelector.setEnabled(False)
                        
                        x_data = dss[0][()]
                        y_data = dss[1][()][self.VTraceXNum]
                        if err_url:
                            err_data = dss[2][()][self.VTraceXNum]
                    x_data_len = len(x_data)
                    y_data_len = len(y_data)
                    if x_data_len != y_data_len:
                        if x_data_len > y_data_len:
                            x_data = x_data[:y_data_len]
                        else:
                            y_data = y_data[:x_data_len]
                
                elif y_ds_type == ds_types['box']:
                    self.VTraceXSelector.setEnabled(True)
                    range_maxX = dss[1].shape[0]
                    self.VTraceXSelector.setRange(-1 * range_maxX, range_maxX - 1)
                    self.VTraceXValue.setText(self._getXValueFromTraceNum(dss[1], self.VTraceXNum))
                    self.VTraceYSelector.setEnabled(True)
                    range_maxY = dss[1].shape[1]
                    self.VTraceYSelector.setRange(-1 * range_maxY, range_maxY - 1)
                    self.VTraceYValue.setText(self._getYValueFromTraceNum(dss[1], self.VTraceYNum))
                    
                    x_data = dss[0][()]
                    y_data = dss[1][()][self.VTraceXNum, self.VTraceYNum, :]
                    if err_url:
                        err_data = dss[2][()][self.VTraceXNum, self.VTraceYNum, :]
            
            ## This is in our case used so far only for IQ plots. The
            ## functionality derives from this application.
            elif x_ds_type == ds_types['matrix']:
                self.VTraceXSelector.setEnabled(True)
                range_max = np.minimum(dss[0].shape[0], dss[1].shape[0])
                self.VTraceXSelector.setRange(-1 * range_max, range_max - 1)
                self.VTraceXValue.setText(self._getXValueFromTraceNum(dss[1], self.VTraceXNum))
                self.VTraceYSelector.setEnabled(False)
                
                x_data = dss[0][()][self.VTraceXNum]
                y_data = dss[1][()][self.VTraceXNum]
            
            elif x_ds_type == ds_types['box']:
                self.VTraceXSelector.setEnabled(True)
                range_maxX = dss[1].shape[0]
                self.VTraceXSelector.setRange(-1 * range_maxX, range_maxX - 1)
                self.VTraceXValue.setText(self._getXValueFromTraceNum(dss[1], self.VTraceXNum))
                self.VTraceYSelector.setEnabled(True)
                range_maxY = dss[1].shape[1]
                self.VTraceYSelector.setRange(-1 * range_maxY, range_maxY - 1)
                self.VTraceYValue.setText(self._getYValueFromTraceNum(dss[1], self.VTraceYNum))
                
                x_data = dss[0][()][self.VTraceXNum, self.VTraceYNum, :]
                y_data = dss[1][()][self.VTraceXNum, self.VTraceYNum, :]
            
            else:
                return
            
            ## Any data manipulation (dB <-> lin scale, etc) is done here
            x_data, y_data, names[0], names[1], units[0], units[1] = _do_data_manipulation(x_data, y_data, names[0], names[1], units[0], units[1], ds_types['vector'], self.manipulation, self.manipulations)

            _axis_timestamp_formatting(graphicsView, x_data, units[0], names[0], "bottom")
            _axis_timestamp_formatting(graphicsView, y_data, units[1],names[1], "left")
            
            # this allows to set a couple of plot related settings
            if view_params:
                aspect = view_params.pop('aspect', False)
                if aspect:
                    graphicsView.setAspectLocked(lock=True, ratio=aspect)
                    # bgcolor = view_params.pop('bgcolor',False)
                    # if bgcolor:
                    #    print tuple(bgcolor)
                    #    graphicsView.setBackgroundColor(tuple(bgcolor))
                
            
            try:
                graphicsView.plotItem.legend.removeItem(names[1])
            except:
                pass

            if not self.user_setting_changed:
                self.plot_style = view_params.get('plot_style', 0)
            self.markersize = view_params.get('markersize', 5)
            try:
                linecolor = view_params['linecolors'][i]
            except (KeyError,IndexError):
                linecolor = (i,4)
            if self.plot_style == self.plot_styles['line']:
                graphicsView.plot(y=y_data, x=x_data, pen=linecolor, name=names[1], connect='finite')
            elif self.plot_style == self.plot_styles['linepoint']:
                symbols = ['+', 'o', 's', 't', 'd']
                graphicsView.plot(y=y_data, x=x_data, pen=linecolor, name=names[1], connect='finite', symbol=symbols[i % len(symbols)], symbolSize=self.markersize)
            elif self.plot_style == self.plot_styles['point']:
                symbols = ['+', 'o', 's', 'd', 't']
                graphicsView.plot(y=y_data, x=x_data, name=names[1], pen=None, symbol=symbols[i % len(symbols)], symbolSize=self.markersize)
            if err_url:
                err = pg.ErrorBarItem(x=x_data, y=y_data, height=err_data, beam=0.25 * scales[0][0])
                graphicsView.getPlotItem().addItem(err)

    # optionally take provided x_name, y_name as labels
    if view_params.get("labels", False):
        graphicsView.setLabel('bottom', view_params['labels'][0], units=units[0])
        graphicsView.setLabel('left', view_params['labels'][1], units=units[1])
    
    plIt = graphicsView.getPlotItem()
    plVi = plIt.getViewBox()
    
    self._last_x_pos = 0
    
    def mouseMoved(mpos):
        mpos = mpos[0]
        if plIt.sceneBoundingRect().contains(mpos):
            mousePoint = plVi.mapSceneToView(mpos)
            xval = mousePoint.x()
            yval = mousePoint.y()
            if plIt.ctrl.logXCheck.isChecked():  # logarithmic scale on x axis
                xval = 10**xval
            if plIt.ctrl.logYCheck.isChecked():  # logarithmic scale on x axis
                yval = 10 ** yval
            self.PointX.setText("X: %.6e %s" % (xval, units[0]))
            self.PointY.setText("Y: %.6e %s" % (yval, units[1]))
            
            try:
                self.data_coord = "%e\t%e\t%e\t%e" % (xval, yval, self._last_x_pos - xval, xval / (self._last_x_pos - xval))
            except ZeroDivisionError:
                pass
            
            self._last_x_pos = xval
    
    self.proxy = pg.SignalProxy(plVi.scene().sigMouseMoved, rateLimit=15, slot=mouseMoved)


def _display_1D_data(self, graphicsView):
    """displays 1d data from the respective dataset.

    This is the most basic way to display data. It plots the values in the 
    dataset against the highest coordinate, ie a VNA trace or the latest IV 
    curve (data vs bias).
    Depending on the ds-type, the QComboBox handles the index of the shown 
    data.
    
    Args:
        self: Object of the PlotWindow class.
        graphicsView: Object of pyqtgraph's PlotWidget class.

    Returns:
        No return variable. The function operates on an object of the 
        PlotWindow class.
    """
    if self.ds_type == ds_types['vector'] or (self.ds_type == -1 and len(self.ds.shape) == 1):  # last expresson is for old hdf-files
        dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ['x_ds_url'])
        
        # timestamps do (not?) have a x_ds_url in the 1d case. This is more a bug to be fixed in the
        # timstamp_ds part of qkit the resulting error is fixed here for now.
        try:
            x_data = dss[0][()][:dss[1].shape[-1]]  # x_data gets truncated to y_data shape if necessarry
        except:
            x_data = [i for i in range(dss[1].shape[-1])]
            units[0] = "#"
            names[0] = "data point number"
        y_data = dss[1][()]
    
    elif self.ds_type == ds_types['coordinate']:
        ## a coordinate does not have any coordinate. it gets plotted against the entry index.
        ## this does not quite work with the unified readout.
        names, units, scales = ['index'], ['#'], [(0, 1)]
        dss, n, u, s = _get_all_ds_names_units_scales(self.ds)
        names.append(n[0])
        units.append(u[0])
        scales.append(s[0])
        # timestamps do (not?) have a x_ds_url in the 1d case. This is more a bug to be fixed in the
        # timstamp_ds part of qkit the resulting error is fixed here for now.
        x_data = [i for i in range(dss[0].shape[-1])]
        y_data = dss[0][()]
    
    elif self.ds_type == ds_types['matrix'] or (self.ds_type == -1 and len(self.ds.shape) == 2):  # last expresson is for old hdf-files
        """
        For a matrix type the data to be displayed on the x-axis depends on the selected plot_type
        """
        if self.PlotTypeSelector.currentIndex() == 1:  # y_ds on x-axis
            dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ['y_ds_url'])
            self.TraceXSelector.setRange(-1 * self.ds.shape[0], self.ds.shape[0] - 1)
            if self.TraceXValueChanged:
                """
                If the trace to be displayed has been changed, the correct dataslice and the displayed
                text has to be adjusted.
                """
                # calc trace number from entered value
                (x0, dx) = _get_axis_scale(_get_ds(self.ds, _get_ds_url(self.ds, 'x_ds_url')))
                num = int((self._traceX_value - x0) / dx)
                self.TraceXNum = num
                self.TraceXSelector.setValue(self.TraceXNum)
                self.TraceXValueChanged = False
            
            y_data = dss[1][()][self.TraceXNum]
            x_data = dss[0][()][:dss[1].shape[-1]]  # x_data gets truncated to y_data shape if neccessary
        
        if self.PlotTypeSelector.currentIndex() == 2:  # x_ds on x-axis
            dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ['x_ds_url'])
            self.TraceXSelector.setRange(-1 * self.ds.shape[1], self.ds.shape[1] - 1)
            if self.TraceYValueChanged:
                """
                If the trace to be displayed has been changed, the correct dataslice and the displayed
                text has to be adjusted.
                """
                
                # calc trace number from entered value
                (x0, dx) = _get_axis_scale(_get_ds(self.ds, _get_ds_url(self.ds, 'y_ds_url')))
                num = int((self._traceY_value - x0) / dx)
                self.TraceYNum = num
                self.TraceYSelector.setValue(self.TraceYNum)
                self.TraceYValueChanged = False
            
            y_data = dss[1][()][:, self.TraceYNum]
            x_data = dss[0][()][:dss[1].shape[0]]  # x_data gets truncated to y_data shape if neccessary
        
        self.TraceXValue.setText(self._getXValueFromTraceNum(self.ds, self.TraceXNum))
        self.TraceYValue.setText(self._getYValueFromTraceNum(self.ds, self.TraceYNum))
    
    
    
    elif self.ds_type == ds_types['box']:
        """
        For a box type. Z data is displayed along the chosen axes in the viewer
        """
        self.TraceXSelector.setRange(-1 * self.ds.shape[0], self.ds.shape[0] - 1)
        self.TraceYSelector.setRange(-1 * self.ds.shape[1], self.ds.shape[1] - 1)
        self.TraceZSelector.setRange(-1 * self.ds.shape[2], self.ds.shape[2] - 1)

        if self.TraceXValueChanged:
            """
            If the trace to be displayed has been changed, the correct dataslice and the displayed
            text has to be adjusted.
            """
            # calc trace number from entered value
            (x0, dx) = _get_axis_scale(_get_ds(self.ds, _get_ds_url(self.ds, 'x_ds_url')))
            num = int((self._traceX_value - x0) / dx) + 1
            self.TraceXNum = num
            self.TraceXSelector.setValue(self.TraceXNum)
            self.TraceXValueChanged = False
        
        if self.TraceYValueChanged:
            """
            If the trace to be displayed has been changed, the correct dataslice and the displayed
            text has to be adjusted.
            """
            # calc trace number from entered value
            (y0, dy) = _get_axis_scale(_get_ds(self.ds, _get_ds_url(self.ds, 'y_ds_url')))
            num = int((self._traceY_value - y0) / dy)
            self.TraceYNum = num
            self.TraceYSelector.setValue(self.TraceYNum)
            self.TraceYValueChanged = False

        if self.TraceZValueChanged:
            """
            If the trace to be displayed has been changed, the correct dataslice and the displayed
            text has to be adjusted.
            """
            # calc trace number from entered value
            (z0, dz) = _get_axis_scale(_get_ds(self.ds, _get_ds_url(self.ds, 'z_ds_url')))
            num = int((self._traceZ_value - z0) / dz)
            self.TraceZNum = num
            self.TraceZSelector.setValue(self.TraceZNum)
            self.TraceZValueChanged = False
        
        self.TraceXValue.setText(self._getXValueFromTraceNum(self.ds, self.TraceXNum))
        self.TraceYValue.setText(self._getYValueFromTraceNum(self.ds, self.TraceYNum))
        self.TraceZValue.setText(self._getZValueFromTraceNum(self.ds, self.TraceZNum))

        if self.PlotTypeSelector.currentIndex() == 5:
            dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ['z_ds_url'])
            x_data = dss[0][()][:dss[1].shape[2]]  # x_data gets truncated to y_data shape if neccessary
            y_data = dss[1][()][self.TraceXNum, self.TraceYNum, :]
        if self.PlotTypeSelector.currentIndex() == 4:
            dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ['y_ds_url'])
            x_data = dss[0][()][:dss[1].shape[1]]
            y_data = dss[1][()][self.TraceXNum, :, self.TraceZNum]
        if self.PlotTypeSelector.currentIndex() == 3:
            dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ['x_ds_url'])
            x_data = dss[0][()][:dss[1].shape[0]]
            y_data = dss[1][()][:, self.TraceYNum, self.TraceZNum]

    
    ## Any data manipulation (dB <-> lin scale, etc) is done here
    x_data, y_data, names[0], names[1], units[0], units[1] = _do_data_manipulation(x_data,
                                                                                   y_data, names[0], names[1], units[0],
                                                                                   units[1], ds_types['vector'],
                                                                                   self.manipulation, self.manipulations)

    if _axis_timestamp_formatting(graphicsView, x_data, units[0], names[0], "bottom") or \
            _axis_timestamp_formatting(graphicsView, y_data, units[1], names[1],"left"):
        if not self.user_setting_changed:
            self.plot_style = self.plot_styles['linepoint']
    
    # if only one entry in the dataset --> point-style
    elif y_data.shape[-1] == 1:
        self.plot_style = self.plot_styles['point']

    if self.plot_style == self.plot_styles['line']:
        graphicsView.plot(y=y_data, x=x_data, clear=True, pen=(200, 200, 100), connect='finite')
        self.linestyle_selector.line.setChecked(True)
    elif self.plot_style == self.plot_styles['linepoint']:
        graphicsView.plot(y=y_data, x=x_data, clear=True, pen=(200, 200, 100), connect='finite', symbol='+')
        self.linestyle_selector.pointLine.setChecked(True)
    elif self.plot_style == self.plot_styles['point']:
        graphicsView.plot(y=y_data, x=x_data, clear=True, pen=None, symbol='+')
        self.linestyle_selector.point.setChecked(True)
    
    plIt = graphicsView.getPlotItem()
    plVi = plIt.getViewBox()
    
    self._last_x_pos = 0
    
    def mouseMoved(mpos):
        mpos = mpos[0]
        if plIt.sceneBoundingRect().contains(mpos):
            mousePoint = plVi.mapSceneToView(mpos)
            xval = mousePoint.x()
            yval = mousePoint.y()
            if plIt.ctrl.logXCheck.isChecked():  # logarithmic scale on x axis
                xval = 10**xval
            if plIt.ctrl.logYCheck.isChecked():  # logarithmic scale on x axis
                yval = 10 ** yval
            self.PointX.setText("X: %.6e %s" % (xval, units[0]))
            self.PointY.setText("Y: %.6e %s" % (yval, units[1]))
            
            try:
                self.data_coord = "%e\t%e\t%e\t%e" % (xval, yval, self._last_x_pos - xval, xval / (self._last_x_pos - xval))
            except ZeroDivisionError:
                pass
            
            self._last_x_pos = xval
            if self.distance_measure[0] == 1: # a ROI is open
                _, x0, y0, [roi, tx, ty,td] = self.distance_measure
                roi.setSize((xval - x0, yval - y0))
                tx.setText("%s" % pg.siFormat(xval - x0, suffix=units[0]))
                tx.setPos((xval + x0) / 2, min(yval, y0))
                ty.setText("%s" % pg.siFormat(-y0 + yval, suffix=units[1]))
                ty.setPos(max(xval, x0), (yval + y0) / 2)
                try:
                    sx, Sx = pg.siScale((xval - x0))
                    td.setText("%s/%s%s" % (pg.siFormat((yval - y0) / (xval - x0) / sx, suffix=units[1]), Sx, units[0]))
                except ZeroDivisionError:
                    td.setText("--")
                td.setPos(min(xval, x0), max(yval, y0))

    self.proxy = pg.SignalProxy(plVi.scene().sigMouseMoved, rateLimit=15, slot=mouseMoved)
    
    def middleClick(mce):
        mce = mce[0]
        if mce.button() == 4:
            mce.accept()
            mousePoint = plVi.mapSceneToView(mce.scenePos())
            xval = mousePoint.x()
            yval = mousePoint.y()
            if plIt.ctrl.logXCheck.isChecked():  # logarithmic scale on x axis
                xval = 10 ** xval
            if plIt.ctrl.logYCheck.isChecked():  # logarithmic scale on x axis
                yval = 10 ** yval
            
            if self.distance_measure[0] is False:
                roi = pg.RectROI((xval, yval), (0, 0))
                tx = pg.TextItem("")
                tx.setPos(xval, yval)
                ty = pg.TextItem("")
                ty.setPos(xval, yval)
                td = pg.TextItem("")
                td.setPos(xval, yval)
                self.distance_measure = [1, xval, yval, [roi, tx, ty,td]]
                [plVi.addItem(i) for i in self.distance_measure[-1]]
            elif self.distance_measure[0] == 1:
                self.distance_measure[0] = 2 # hold distance measure ROI
            else:
                if len(self.distance_measure) >= 4: # remove ROI
                    [plVi.removeItem(i) for i in self.distance_measure[-1]]
                self.distance_measure[0] = False

    self.distance_measure = self.__dict__.get('distance_measure', [False])

    self.proxy2 = pg.SignalProxy(plVi.scene().sigMouseClicked,slot=middleClick)


def _display_2D_data(self, graphicsView):
    """displays a 2d matrix of data color coded.

    The 2d color plot gets treated similar: data and labels are looked up with 
    respect to the dataset type and the PlotType for ds-boxes.
    Here the only relevant information is the data to be plotted (a 2d matrix 
    of values) and the x- and y-axis. From them the specific "slice" of data is
    selected and the lable infos are read out. The exact data in the datasets 
    on the x- and y-axis are not needed. The underlying pyqtgraph fct works 
    with a position (x0, y0) and a scale (dx, dy) for creating the axis tics 
    and tic lables. This means it is not possible to display any non-linear 
    scaled axis data.
    Depending on the ds-type, the QComboBox handles the index of the shown 
    data.
    
    Args:
        self: Object of the PlotWindow class.
        graphicsView: Modified object of pyqtgraph's ImageView class.

    Returns:
        No return variable. The function operates on an object of the 
        PlotWindow class.
    """
    
    if self.ds_type == ds_types['matrix']:
        """
        The matrix ds-type only knows one 2d plotting option. x_ds on x- and y_ds on y-axis
        """
        dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ['x_ds_url', 'y_ds_url'])
        try:
          data = dss[2][()]
        except IOError as e:
              print("Could not open data file")
              print(e)
              return
        
        fill_x = dss[2].shape[0]
        fill_y = dss[2].shape[1]
        self.TraceXValue.setText(self._getXValueFromTraceNum(self.ds, self.TraceXNum))
        self.TraceYValue.setText(self._getYValueFromTraceNum(self.ds, self.TraceYNum))
    if self.ds_type == ds_types['box']:
        """
        The box ds-type can be plotted from 3 different "viewing directions" (PlotType). Depending on its 
        setting the x- and y-axis are set.
        The ds-type box also has a z_ds_url.
        """
        self.TraceXSelector.setRange(-1 * self.ds.shape[0], self.ds.shape[0] - 1)
        self.TraceYSelector.setRange(-1 * self.ds.shape[-1], self.ds.shape[-1] - 1)
        if self.PlotTypeSelector.currentIndex() == 0:  # y_ds on x-axis; z_ds on y-axis
            if self.TraceXValueChanged:
                # calc trace number from entered value
                num = int((self._traceX_value - self.ds.attrs.get("x0", 0)) / (self.ds.attrs.get("dx", 1)))
                self.TraceXNum = num
                self.TraceXSelector.setValue(self.TraceXNum)
                self.TraceXValueChanged = False
            
            dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ['y_ds_url', 'z_ds_url'])
            try:
              data = dss[2][()][self.TraceXNum, :, :]
            except IOError as e:
              print("Could not open data file")
              print(e)
              return
            
            fill_x = dss[2].shape[1]
            fill_y = dss[2].shape[2]
        
        if self.PlotTypeSelector.currentIndex() == 1:  # x_ds on x-axis; z_ds on y-axis
            if self.TraceYValueChanged:
                # calc trace number from entered value
                num = int((self._traceY_value - self.ds.attrs.get("y0", 0)) / (self.ds.attrs.get("dy", 1)))
                self.TraceYNum = num
                self.TraceYSelector.setValue(self.TraceYNum)
                self.TraceYValueChanged = False
            
            dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ['x_ds_url', 'z_ds_url'])
            try:
              data = dss[2][()][:, self.TraceYNum, :]
            except IOError as e:
              print("Could not open data file")
              print(e)
              return
            
            fill_x = dss[2].shape[0]
            fill_y = dss[2].shape[2]
        
        if self.PlotTypeSelector.currentIndex() == 2:  # x_ds on x-axis; y_ds on y-axis
            if self.TraceZValueChanged:
                # calc trace number from entered value
                num = int((self._traceZ_value - self.ds.attrs.get("z0", 0)) / (self.ds.attrs.get("dz", 1)))
                self.TraceZNum = num
                self.TraceZSelector.setValue(self.TraceZNum)
                self.TraceZValueChanged = False
            
            dss, names, units, scales = _get_all_ds_names_units_scales(self.ds, ['x_ds_url', 'y_ds_url'])
            try:
              data = dss[2][()][:, :, self.TraceZNum]
            except IOError as e:
              print("Could not open data file")
              print(e)
              return
            
            fill_x = dss[2].shape[0]
            fill_y = dss[2].shape[1]
        
        self.TraceXValue.setText(self._getXValueFromTraceNum(self.ds, self.TraceXNum))
        self.TraceYValue.setText(self._getYValueFromTraceNum(self.ds, self.TraceYNum))
        self.TraceZValue.setText(self._getZValueFromTraceNum(self.ds, self.TraceZNum))

    _, data, _, _, _, units[2] = _do_data_manipulation(None, data, None, None, None, units[2], ds_types['vector'], self.manipulation, self.manipulations, colorplot=True)

    graphicsView.clear()
    graphicsView.view.setLabel('left', names[1], units=units[1])
    graphicsView.view.setLabel('bottom', names[0], units=units[0])
    graphicsView.view.setTitle(names[2] + " (" + units[2] + ")")
    graphicsView.view.invertY(False)
    
    # pos is the zero-point of the axis  
    # scale is responsible for the "accidential" correct display of the axis
    # for downsweeps scale has negative values and extends the axis from the min values into the correct direction
    if np.all(np.isnan(data)):
        data[(0,) * len(data.shape)] = 0
        print("Your Data array is all NaN. I set the first value to not blow up graphics window.")
    graphicsView.setImage(data, pos=(scales[0][0] - scales[0][1] / 2., scales[1][0] - scales[1][1] / 2.), scale=(scales[0][1], scales[1][1]))
    graphicsView.show()
    
    # Fixme roi ...
    """
    The error in the ROI may come from a not always correct setting of xmin/xmax...
    """
    graphicsView.roi.setSize([abs(fill_x * scales[0][1]), abs(fill_y * scales[1][1])])
    graphicsView.roi.setPos([scales[0][0], scales[1][0]])
    graphicsView.roi.setAcceptedMouseButtons(Qt.RightButton)
    graphicsView.roi.sigClicked.connect(lambda: self.clickRoi(graphicsView.roi.pos(), graphicsView.roi.size()))
    
    imIt = graphicsView.getImageItem()
    imVi = graphicsView.getView()
    
    def mouseMoved(mpos):
        mpos = mpos[0]
        mousePoint = imIt.mapFromScene(mpos)
        x_index = int(mousePoint.x())
        y_index = int(mousePoint.y())
        
        xval = scales[0][0] + x_index * scales[0][1]
        yval = scales[1][0] + y_index * scales[1][1]
        if 0<= x_index < fill_x and 0 <= y_index < fill_y:
            zval = data[x_index][y_index]
        else:
            zval = 0
        self.PointX.setText("X: %.6e %s" % (xval, units[0]))
        self.PointY.setText("Y: %.6e %s" % (yval, units[1]))
        self.PointZ.setText("Z: %.6e %s" % (zval, units[2]))
        self.data_coord = "%g\t%g\t%g" % (xval, yval, zval)

        if self.distance_measure[0] == 1: # a ROI is open
            _, x0, y0, [roi, tx, ty, td] = self.distance_measure
            roi.setSize((xval - x0, yval - y0))
            tx.setText("%s" % pg.siFormat(xval - x0, suffix=units[0]))
            tx.setPos((xval + x0) / 2, min(yval, y0))
            ty.setText("%s" % pg.siFormat(-y0 + yval, suffix=units[1]))
            ty.setPos(max(xval, x0), (yval + y0) / 2)
            try:
                sx,Sx = pg.siScale((xval - x0))
                td.setText("%s/%s%s" % (pg.siFormat((yval - y0) / (xval - x0)/sx, suffix=units[1]), Sx, units[0]))
            except ZeroDivisionError:
                td.setText("--")
            td.setPos(min(xval, x0), max(yval, y0))
        
    self.proxy = pg.SignalProxy(imVi.scene().sigMouseMoved, rateLimit=15, slot=mouseMoved)
    def middleClick(mce):
        mce = mce[0]
        if mce.button() == 4:
            mce.accept()
            mousePoint = imIt.mapFromScene(mce.scenePos())
            xval = scales[0][0] + int(mousePoint.x()) * scales[0][1]
            yval = scales[1][0] + int(mousePoint.y()) * scales[1][1]

            if self.distance_measure[0] is False:
                roi = pg.RectROI((xval, yval), (0,0))
                tx = pg.TextItem("")
                tx.setPos(xval, yval)
                ty = pg.TextItem("")
                ty.setPos(xval,yval)
                td = pg.TextItem("")
                td.setPos(xval, yval)
                self.distance_measure = [1,xval,yval,[roi, tx, ty, td]]
                [imVi.addItem(i) for i in self.distance_measure[-1]]
            elif self.distance_measure[0] == 1:
                self.distance_measure[0] = 2  # hold distance measure ROI
            else:
                if len(self.distance_measure) >= 4:  # remove ROI
                    [imVi.removeItem(i) for i in self.distance_measure[-1]]
                self.distance_measure[0] = False
                
    self.distance_measure = self.__dict__.get('distance_measure',[False])
    self.proxy2 = pg.SignalProxy(imVi.scene().sigMouseClicked,slot=middleClick)


def _display_table(self, graphicsView):
    """displays the data values in a table.

    This function shows a table with the numerical values of the dataset with
    the shape depending on the ds-type.
    
    Args:
        self: Object of the PlotWindow class.
        graphicsView: Object of pyqtgraph's TableWidget class.

    Returns:
        No return variable. The function operates on an object of the 
        PlotWindow class.
    """
    ## ds-type "box" is not (yet?) implemented here. This may be done in the 
    ## future.
    data = np.array(self.ds)
    if self.ds_type == ds_types['matrix']:
        data = data.transpose()
    if self.ds_type == ds_types['vector'] or self.ds_type == ds_types['coordinate']:
        data_tmp = np.empty((1, data.shape[0]), dtype=np.float64)
        data_tmp[0] = data
        data = data_tmp.transpose()
    if self.ds_type == ds_types["txt"]:
        data_tmp = []
        for d in data:
            data_tmp.append([d])
        data = np.array(data_tmp)
        graphicsView.setFormat(str(data))
    graphicsView.setData(data)


def _display_text(self, graphicsView):
    """displays the data (=String) in the dataset.

    This function mainly handles our JSON encoded entries in the settings 
    datasets. The JSON dict is read out and displayed with indentations to be
    nicely readable. If the input string is not JSON encodable it is displayed
    by the _display_string() fcn as it is.
    
    Args:
        self: Object of the PlotWindow class.
        graphicsView: Object of PyQt's QPlainTextEdit class.

    Returns:
        No return variable. The function operates on an object of the 
        PlotWindow class.
    """
    try:
        json_dict = json.loads(self.ds[()][0])
    except ValueError:
        txt = _display_string(self.ds)
    else:
        txt = pprint.pformat(json_dict, indent=4)
        txt = txt.replace("u'",'').replace("'","").replace("{"," ").replace("}","").replace(",\n","\n")
    graphicsView.insertPlainText(txt.rstrip())


def _display_string(ds):
    """Reads the sting in dataset ds and returns a formatted string.

    Args:
        ds: hdf_dataset.

    Returns:
        Formatted string.
    """
    data = np.array(ds)
    txt = ""
    for d in data:
        txt += d + '\n'
    return txt


""" A few handy methods for label and scale """


def _get_ds_url(ds, url):
    """Gets the absolute ds_url.

    If 'url' is already a absolute url in the h5 file it gets returned 
    unchanged.
    Otherwise the absolute url in read out of the ds-metadata.
    
    Args:
        ds: hdf_dataset.
        url: string of absolute or relative ds_url

    Returns:
        String of the absolute ds_url.
    """
    
    if _get_ds(ds, url):
        return url
    else:
        return ds.attrs.get(url, '_none_')


def _get_ds(ds, ds_url):
    """Returns the dataset associated with the given ds_url in the same file as
    the given dataset ds.
    
    Args:
        ds: hdf_dataset.
        ds_url: absolute ds_url

    Returns:
        Object of hdf_dataset class.
    """
    try:
        return ds.file[ds_url]
    except:
        return None


def _get_axis_scale(ds):
    """Returns the scale of a coordinate, x0 and dx at an assumed linear 
    scaling.
    
    Args:
        ds: hdf_dataset.

    Returns:
        Tuple with of axis origin and delta distance.
    """
    try:
        x0 = ds[0]
        dx = ds[1] - ds[0]
        return (x0, dx)
    except:
        return (0, 1)


def _get_unit(ds):
    """Returns the unit of a dataset.
    
    Args:
        ds: hdf_dataset.

    Returns:
        String with unit.
    """
    try:
        return str3(ds.attrs.get('unit', b'_none_'))
    except AttributeError as e:
        #print(ds)
        print("Qviewkit _get_unit:",e)
        return '_none_'

def _get_name(ds):
    """Returns the name of a dataset.

    Args:
        ds: hdf_dataset.

    Returns:
        String with name.
    """
    if ds is None:
        return '_none_'
    try:
        return str3(ds.attrs.get('name', b'_none_'))
    except AttributeError as e:
        #print(ds)
        print("Qviewkit _get_name:",e)
        return '_none_'


def _get_all_ds_names_units_scales(ds, ds_urls=[]):
    """Reads and returns all the relevant information of the given datasets.

    This function gathers all the needed metadata to correctly display the
    dataset given in ds. The metadata of the datasets associated with the given
    urls are read out (data, name, unit, and scale). This information is used
    by the _display_...() fcts to correcly label and scale the plots. The order
    in the input- and output-list is maintained.
    
    Args:
        ds: hdf_dataset.
        ds_url: List of urls to be read. 

    Returns:
        Multiple lists containing the data/metadata 'dataset', 'name', 'unit',
        and 'scale' of the input urls.
    """
    dss = []
    
    for ds_url in ds_urls:
        dss.append(_get_ds(ds, _get_ds_url(ds, ds_url)))
    # the last dataset is always the displayed data.
    dss.append(ds)
    
    names = []
    units = []
    scales = []
    for ds in dss:
        if ds is None:
            names.append('_none_')
            units.append('_none_')
            scales.append('_none_')
        else:
            names.append(_get_name(ds))
            units.append(_get_unit(ds))
            scales.append(_get_axis_scale(ds))
    return dss, names, units, scales


""" Unify the data manipulation to share code """


def _do_data_manipulation(x_data, y_data, x_name, y_name, x_unit, y_unit, ds_type, manipulation, manipulations, colorplot=False):
    """Data manipulation for display gets done here.
    
    This function gathers all the needed metadata to correctly display the
    dataset given in ds. The metadata of the datasets associated with the given 
    urls are read out (y_data, name, unit, and scale). This information is used
    by the _display_...() fcts to correcly label and scale the plots. The order
    in the input- and output-list is maintained.
    
    Args:
        x_data: Numpy array of the to be displayed / to be manipulated data.
            This needs to be changed for histograms.
        y_data: Numpy array of the to be displayed / to be manipulated data.
        ds_type: Dataset type for information about the dimension.
        manipulation: Integer, power of 2 number indicating the manipulation to
            be performed.
        manipulations: Lookup dict, mapping a manipulation to a power of 2.

    Returns:
        Tuple of a numpy array containing the manipulated data with the same
        dimension as the input array and a string of the data unit after the 
        manipulation.
    """
    # set the y data  to the decibel scale
    
    # unwrap the phase
    if manipulation & manipulations['wrap']:
        y_data[~np.isnan(y_data)] = np.unwrap(y_data[~np.isnan(y_data)])
    
    if manipulation & manipulations['linear']:
        if len(y_data.shape) == 1:
            y_data = y_data - np.nan_to_num(np.linspace(y_data[0], y_data[-1], len(y_data)))
        elif len(y_data.shape) == 2:
            y_data = y_data - np.outer(y_data[:, -1] - y_data[:, 0], np.linspace(0, 1, y_data.shape[1]))
        else:
            print("linear correction not implemented for %iD" % len(y_data.shape))
        
        if colorplot:
            ## This manipulation removes all zeros which would blow up the color scale.
            ## Only relevant for matrices and boxes in 2d
            if manipulation & manipulations['remove_zeros']:
                y_data[np.where(y_data == 0)] = np.NaN  # replace all exact zeros in the hd5 y_data with NaNs, otherwise the 0s in uncompleted files blow up the colorscale

    # subtract mean value (offset correction) from the data along the x-axis
    if manipulation & manipulations['sub_offset_avg_x']:
        # ignore division by zero
        old_warn = np.seterr(divide='print')
        np.seterr(**old_warn)
        y_data = y_data - np.nanmean(y_data, axis=0, keepdims=True)

    # subtract mean value (offset correction) from the data along the y-axis
    if manipulation & manipulations['sub_offset_avg_y']:
        # ignore division by zero
        old_warn = np.seterr(divide='print')
        np.seterr(**old_warn)
        y_data = y_data - np.nanmean(y_data, axis=1, keepdims=True)

    # divide data by mean value (normalization) along the x-axis
    if manipulation & manipulations['norm_data_avg_x']:
        # ignore division by zero
        old_warn = np.seterr(divide='print')
        np.seterr(**old_warn)
        y_data = y_data / np.nanmean(y_data, axis=0, keepdims=True)

    # divide data by mean value (normalization) along the y-axis
    if manipulation & manipulations['norm_data_avg_y']:
        # ignore division by zero
        old_warn = np.seterr(divide='print')
        np.seterr(**old_warn)
        y_data = y_data / np.nanmean(y_data, axis=1, keepdims=True)

    if manipulation & manipulations['dB']:
        y_data = 20 * np.log10(y_data)
        y_unit = 'dB'

    if manipulation & manipulations['histogram']:
        if y_data.ndim == 1:
            y_data, x_data = np.histogram(y_data[~np.isnan(y_data)], bins='auto')
            x_data = x_data[:-1]+(x_data[1]-x_data[0])/2.
            x_name = y_name
            x_unit = y_unit
            y_name = 'counts'
            y_unit = ''
        elif y_data.ndim == 2:
            print('2D histogram not yet impemented')  # FIXME: np.histogram2d (MMW)

    
    return x_data, y_data, x_name, y_name, x_unit, y_unit

def _axis_timestamp_formatting(graphicsView,data,unit,name,orientation):
    if unit=="s" and data[tuple(-1 for x in data.shape)] > 9.4e8: # about the year 2000 in unixtime
        from .pg_time_axis import DateAxisItem
        if type(graphicsView.getPlotItem().axes[orientation]['item']) != DateAxisItem:
            axis = DateAxisItem(orientation=orientation)
            axis.attachToPlotItem(graphicsView.getPlotItem())
            axis.enableAutoSIPrefix(False)
        graphicsView.setLabel(orientation, name)
        return True
    else:
        graphicsView.setLabel(orientation, name, units=unit)
        return False



# -*- coding: utf-8 -*-
"""
@author: MiWi,HR,S1@KIT/2017,2018
"""
import qkit
import logging
from pkgutil import find_loader


def _load_visa():
    try:
        try:
            import pyvisa as visa
        except:
            import visa
    except Exception as e:
        qkit.cfg['load_visa'] = False
        raise type(e)('Failed loading visa. Check if you have NI VISA or pyvisa-py installed. Original error: ' + str(e))
    else:
        from pkg_resources import get_distribution
        from distutils.version import LooseVersion
        if LooseVersion(get_distribution('pyvisa').version) < LooseVersion("1.5.0"):
            logging.warning("Old pyvisa version loaded. Please update to a version > 1.5.x")
            # compatibility with old visa lib
            qkit.visa = visa
            qkit.visa.__version__ = get_distribution('pyvisa').version
            qkit.visa.qkit_visa_version = 1 #This makes it just much easier to distinguish between the main versions
        else:
            # active py visa version
            logging.info("Modern pyvisa version loaded. Version %s" % visa.__version__)
            try:
                rm = visa.ResourceManager(qkit.cfg.get('visa_backend',""))
                qkit.visa = rm
                qkit.visa.__version__ = visa.__version__
                qkit.visa.qkit_visa_version = 2
                qkit.visa.VisaIOError = visa.VisaIOError
                
                def instrument(resource_name, **kwargs):
                    return rm.open_resource(resource_name, **kwargs)
                qkit.visa.instrument = instrument
                # define data types:
                qkit.visa.double = "d"
                qkit.visa.single = "f"
                qkit.visa.dtypes = {1:qkit.visa.single,
                                    3:qkit.visa.double,
                                    "d":"d","f":"f"}
            except OSError:
                raise OSError('Failed creating ResourceManager. Check if you have NI VISA or pyvisa-py installed.')

class DummyVisa(object):
    def __getattr__(self,name):
        raise qkit.QkitCfgError("Please set qkit.cfg['load_visa'] = True if you need visa.")

if qkit.cfg.get('load_visa',False):
    _load_visa()
else:
    qkit.visa = DummyVisa()

"""
doc snipplet from 
https://pyvisa.readthedocs.io/en/stable/migrating.html
"""


import qkit
from qkit.core.instrument_base import Instrument
import types
import logging
import numpy as np
import time

global conversion_factor
conversion_factor = 52.02477434  # mT/A, solenoid in white fridge


class virtual_pwsMagnet(Instrument):
    '''
    This is the driver for a virtual Power Supply for the solenoid coil

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'Tektronix_PWS4205', address='<USB address>, reset=<bool>')
    USB address e.g. USB0::0x0699::0x0390::C010427::INSTR
    '''

    def __init__(self, name, PWS1='PWS1', PWS2='PWS2'):
        '''
        Initializes a virtual magnet built upon two PWS4205 wired in parallel
        connected to solenoid coil in the white fridge
        both current sources deliver up to 5A, but coil will quench
        lower
        
        Input:
          name (string)    : name of the instrument
          PWS1 (string) : name of the first power source, default=PWS1
          PWS2 (string) : name of the second power source, default=PWS2
        '''
        logging.info(__name__ + ' : Initializing instrument virtual power source')
        Instrument.__init__(self, name, tags=['virtual'])
        try:
            self._PWS1 = qkit.instruments.get(PWS1)
        except:
            logging.error('no PWS1 found ... aborting')
            return
        try:
            self._PWS2 = qkit.instruments.get(PWS2)
        except:
            logging.error('no PWS2 found ... aborting')
            return

        self.add_parameter('current', type=float,
                           flags=Instrument.FLAG_GETSET, units='A')
        self.add_parameter('voltage', type=float,
                           flags=Instrument.FLAG_GET, units='V')
        self.add_parameter('status', type=bool,
                           flags=Instrument.FLAG_GETSET)
        self.add_parameter('magneticField', type=float,
                           flags=Instrument.FLAG_GET, units='mT')

        self.add_function('get_all')
        self.add_function('ramp_current')
        self.add_function('ramp_magenticField')
        self.add_function('check_for_quench')
        self.add_function('on')
        self.add_function('off')
        self.get_all()

    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + 'get_all()')
        self.get('current')
        self.get('voltage')
        self.get('status')
        self.get('magneticField')

    def do_get_current(self):
        '''
        Reads the current signal from the instrument

        Input:
            None

        Output:
            current (float) : current in amps
        '''
        logging.debug(__name__ + 'do_get_current()')
        self._current = getattr(self._PWS1, 'get_current')() + getattr(self._PWS2, 'get_current')()
        return self._current

    def do_set_current(self, curr, wait=0.2):
        '''
        Sets the Amplitude of the signal

        Input:
            curr (float) : current in amps
            wait (float) : waiting time in sec, default=0.2

        Output:
            None
        '''
        if curr < 0:
            logging.error('Negative currents now allowed!')
            return
        logging.debug(__name__ + 'do_set_current()')
        getattr(self._PWS1, 'set_current')(curr / 2.)
        time.sleep(wait)
        getattr(self._PWS2, 'set_current')(curr / 2.)
        time.sleep(wait)
        self._current = self.do_get_current()
        self.get_all()

    def do_get_voltage(self):
        '''
        Reads the voltage signal from the instrument

        Input:
            None

        Output:
            volt (float) : Voltage in Volts
        '''
        logging.debug(__name__ + 'do_get_voltage()')
        self._voltage = (getattr(self._PWS1, 'get_voltage')() + getattr(self._PWS2, 'get_voltage')()) / 2.
        return self._voltage

    def do_get_status(self):
        '''
        Reads the output status from the instrument
        'On' only whene all sub-instruments are on
        Input:
            None

        Output:
            status (boolean)
        '''
        logging.debug(__name__ + 'do_get_status()')
        status1 = bool(getattr(self._PWS1, 'get_status')())
        status2 = bool(getattr(self._PWS2, 'get_status')())
        if status1 and status2:
            self._status = status1
        else:
            self._status = False
            getattr(self._PWS1, 'off')()
            getattr(self._PWS2, 'off')()
        return bool(self._status)

    def do_set_status(self, status):
        '''
        Sets the output status of the instrument with
        additional status check

        Input:
            status (boolean) 

        Output:
            None
        '''
        logging.debug(__name__ + 'do_set_status()')
        getattr(self._PWS1, 'set_status')(status)
        getattr(self._PWS2, 'set_status')(status)
        self._status = self._do_get_statur()

    def do_get_magneticField(self):
        logging.debug(__name__ + 'do_get_magneticField()')
        self._magneticField = (getattr(self._PWS1, 'get_current')() + getattr(self._PWS2,
                                                                              'get_current')()) * conversion_factor
        return self._magneticField

    def ramp_magenticField(self, target, step=conversion_factor * 1e-2, wait=0.2, showvalue=True):
        '''
        Ramps the Amplitude of the signal
        Calls ramp_current with the input values convrted to current units

        Input:
            target (float) : target magentic field in mT
            step (float) : magnetic field steps in mT
            wait (float) : waiting time in sec, default=0.2
            showvalue (bool) : print values, default=True

        Output:
            None
        '''
        logging.debug(__name__ + 'ramp_magneticField()')
        target_current = target / conversion_factor
        step_current = step / conversion_factor
        self.ramp_current(self, target=target_current, step=step_current, wait=wait, showvalue=showvalue,
                          print_mag=True)

    def ramp_current(self, target, step=1e-2, wait=0.2, showvalue=True, print_mag=False):
        '''
        Ramps the Amplitude of the signal

        Input:
            target (float) : target current in amps
            step (float) : current steps in amps, default=1e-2
            wait (float) : waiting time in sec, default=0.2
            showvalue (bool) : print values, default=True
            print_mag (bool) : prints values in magentic field units, default=False

        Output:
            None
        '''
        logging.debug(__name__ + 'ramp_current()')
        voltage = self.do_get_voltage()
        current = self.do_get_current()
        if voltage > 0.95:
            print "WARNING! Magnet quench!! ramping down the coil..."
            self.ramp_current(0., 2e-3, wait=wait, showvalue=showvalue)
            return
        else:
            if target < current: step = -step
            Is = np.arange(current, target + step, step)
            for I in Is:
                self.do_set_current(I, wait)
                """
                if print_mag:
                    value_print = 'B = ' + str(self.do_get_current()*conversion_factor) + 'mT'
                else:
                    value_print = 'I = ' + str(self.do_get_current()) + 'A'
                #if showvalue:
                    #print value_print+ ", V1=" + str(getattr(self._PWS1, 'get_voltage')()) + "V, " + "V2=" + str(getattr(self._PWS2, 'get_voltage')()) + "V"
                    """
        voltage = self.do_get_voltage()
        current = self.do_get_current()
        if target != 0:
            res = int(np.trunc(voltage / current * 1000))
        else:
            res = 0
        """
        if print_mag:
            value_print = 'B= '+str(current*conversion_factor) + 'mT'
        else:
            value_print = 'I= '+str(current) + 'A'
        print "Target value reached: " + value_print+", Voltage is " + str(voltage) + " V, resistance = " + str(res) + " mOhm"
        """

    def check_for_quench(self, wait=5, threshold=0.95, repetitions=1000):
        '''
        Checks the magent for quench

        Input:
            wait (float) : waiting time in sec, default=5
            threshold (float) : maximum voltage in volts, default=0.95
            repetitions (int) : number of times quenching is checked, default=1000

        Output:
            None
        '''
        logging.debug(__name__ + 'check_for_quench()')
        for i in range(repetitions):
            time.sleep(wait)
            voltage = self.do_get_voltage()
            current = self.do_get_current()
            print "V=" + str(np.trunc(voltage * 1000) / 1000.) + "V, I=" + str(
                np.trunc(current * 1000) / 1000.) + "A, R=" + str(int(np.trunc(voltage / current * 1000))) + "mOhm"
            if voltage > threshold:
                print "WARNING! Magnet quench!! ramping down the coil..."
                self.ramp_current(0., 2e-3, wait=0.2, showvalue=True)
                return

    def off(self):
        '''
        Sets status to False

        Input:
            None

        Output:
            None
        '''
        self.set_status(False)

    def on(self):
        '''
        Sets status to True

        Input:
            None

        Output:
            None
        '''
        self.set_status(True)



# nidaq.py, python wrapper for NIDAQ DLL
# Reinier Heeres <reinier@heeres.eu>, 2008
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import ctypes
import types
import numpy
import logging
import time

nidaq = ctypes.windll.nicaiu

int32 = ctypes.c_long
uInt32 = ctypes.c_ulong
uInt64 = ctypes.c_ulonglong
float64 = ctypes.c_double
TaskHandle = uInt32

DAQmx_Val_Cfg_Default = int32(-1)

DAQmx_Val_RSE               = 10083
DAQmx_Val_NRSE              = 10078
DAQmx_Val_Diff              = 10106
DAQmx_Val_PseudoDiff        = 12529

_config_map = {
    'DEFAULT': DAQmx_Val_Cfg_Default,
    'RSE': DAQmx_Val_RSE,
    'NRSE': DAQmx_Val_NRSE,
    'DIFF': DAQmx_Val_Diff,
    'PSEUDODIFF': DAQmx_Val_PseudoDiff,
}

DAQmx_Val_Volts             = 10348
DAQmx_Val_Rising            = 10280
DAQmx_Val_FiniteSamps       = 10178
DAQmx_Val_GroupByChannel    = 0
DAQmx_Val_GroupByScanNumber = 1
DAQmx_Val_ChanPerLine       = 0
DAQmx_Val_ChanForAllLines   = 1

DAQmx_Val_CountUp           = 10128
DAQmx_Val_CountDown         = 10124
DAQmx_Val_ExtControlled     = 10326

def CHK(err):
    '''Error checking routine'''

    if err < 0:
        buf_size = 100
        buf = ctypes.create_string_buffer('\000' * buf_size)
        nidaq.DAQmxGetErrorString(err, ctypes.byref(buf), buf_size)
        raise RuntimeError('Nidaq call failed with error %d: %s' % \
            (err, repr(buf.value)))

def buf_to_list(buf):
    name = ''
    namelist = []
    for ch in buf:
        if ch in '\000 \t\n':
            name = name.rstrip(',')
            if len(name) > 0:
                namelist.append(name)
                name = ''
            if ch == '\000':
                break
        else:
            name += ch

    return namelist

def get_device_names():
    '''Return a list of available NIDAQ devices.'''

    bufsize = 1024
    buf = ctypes.create_string_buffer('\000' * bufsize)
    nidaq.DAQmxGetSysDevNames(ctypes.byref(buf), bufsize)
    return buf_to_list(buf)

def reset_device(dev):
    '''Reset device "dev"'''
    nidaq.DAQmxResetDevice(dev)

def get_physical_input_channels(dev):
    '''Return a list of physical input channels on a device.'''

    bufsize = 1024
    buf = ctypes.create_string_buffer('\000' * bufsize)
    nidaq.DAQmxGetDevAIPhysicalChans(dev, ctypes.byref(buf), bufsize)
    return buf_to_list(buf)

def get_physical_output_channels(dev):
    '''Return a list of physical output channels on a device.'''

    bufsize = 1024
    buf = ctypes.create_string_buffer('\000' * bufsize)
    nidaq.DAQmxGetDevAOPhysicalChans(dev, ctypes.byref(buf), bufsize)
    return buf_to_list(buf)

def get_digital_output_channels(dev):
    '''Return a list of physical output channels on a device.'''

    bufsize = 1024
    buf = ctypes.create_string_buffer('\000' * bufsize)
    nidaq.DAQmxGetDevDOLines(dev, ctypes.byref(buf), bufsize)
    return buf_to_list(buf)

def get_physical_counter_channels(dev):
    '''Return a list of physical counter channels on a device.'''

    bufsize = 1024
    buf = ctypes.create_string_buffer('\000' * bufsize)
    nidaq.DAQmxGetDevCIPhysicalChans(dev, ctypes.byref(buf), bufsize)
    return buf_to_list(buf)

def read(devchan, samples=1, freq=10000.0, minv=-10.0, maxv=10.0,
            timeout=10.0, config=DAQmx_Val_Cfg_Default):
    '''
    Read up to max_samples from a channel. Seems to have trouble reading
    1 sample!

    Input:
        devchan (string): device/channel specifier, such as Dev1/ai0
        samples (int): the number of samples to read
        freq (float): the sampling frequency
        minv (float): the minimum voltage
        maxv (float): the maximum voltage
        timeout (float): the time in seconds to wait for completion
        config (string or int): the configuration of the channel

    Output:
        A numpy.array with the data on success, None on error
    '''

    if type(config) is str:
        if config in _config_map:
            config = _config_map[config]
        else:
            return None
    if type(config) is not int:
        return None

    if samples == 1:
        retsamples = 1
        samples = 2
    else:
        retsamples = samples

    data = numpy.zeros(samples, dtype=numpy.float64)

    taskHandle = TaskHandle(0)
    read = int32()
    try:
        CHK(nidaq.DAQmxCreateTask("", ctypes.byref(taskHandle)))
        CHK(nidaq.DAQmxCreateAIVoltageChan(taskHandle, devchan, "",
            config,
            float64(minv), float64(maxv),
            DAQmx_Val_Volts, None))

        if retsamples > 1:
            CHK(nidaq.DAQmxCfgSampClkTiming(taskHandle, "", float64(freq),
                DAQmx_Val_Rising, DAQmx_Val_FiniteSamps,
                uInt64(samples)));
            CHK(nidaq.DAQmxStartTask(taskHandle))
            CHK(nidaq.DAQmxReadAnalogF64(taskHandle, samples, float64(timeout),
                DAQmx_Val_GroupByChannel, data.ctypes.data,
                samples, ctypes.byref(read), None))
        else:
            CHK(nidaq.DAQmxReadAnalogScalarF64(taskHandle, float64(timeout),
                data.ctypes.data, None))
            read = int32(1)

    except Exception, e:
        logging.error('NI DAQ call failed: %s', str(e))
    finally:
        if taskHandle.value != 0:
            nidaq.DAQmxStopTask(taskHandle)
            nidaq.DAQmxClearTask(taskHandle)

    if read > 0:
        if retsamples == 1:
            return data[0]
        else:
            return data[:read.value]
    else:
        return None

def write(devchan, data, freq=10000.0, minv=-10.0, maxv=10.0,
                timeout=10.0):
    '''
    Write values to channel

    Input:
        devchan (string): device/channel specifier, such as Dev1/ao0
        data (int/float/numpy.array): data to write
        freq (float): the the minimum voltage
        maxv (float): the maximum voltage
        timeout (float): the time in seconds to wait for completion

    Output:
        Number of values written
    '''

    if type(data) in (int, float):
        data = numpy.array([data], dtype=numpy.float64)
    elif isinstance(data, numpy.ndarray):
        if data.dtype is not numpy.float64:
            data = numpy.array(data, dtype=numpy.float64)
    elif len(data) > 0:
        data = numpy.array(data, dtype=numpy.float64)
    samples = len(data)

    taskHandle = TaskHandle(0)
    written = int32()
    try:
        CHK(nidaq.DAQmxCreateTask("", ctypes.byref(taskHandle)))
        CHK(nidaq.DAQmxCreateAOVoltageChan(taskHandle, devchan, "",
            float64(minv), float64(maxv), DAQmx_Val_Volts, None))

        if len(data) == 1:
            CHK(nidaq.DAQmxWriteAnalogScalarF64(taskHandle, 1, float64(timeout),
                float64(data[0]), None))
            written = int32(1)
        else:
            CHK(nidaq.DAQmxCfgSampClkTiming(taskHandle, "", float64(freq),
                DAQmx_Val_Rising, DAQmx_Val_FiniteSamps, uInt64(samples)))
            CHK(nidaq.DAQmxWriteAnalogF64(taskHandle, samples, 0, float64(timeout),
                DAQmx_Val_GroupByChannel, data.ctypes.data,
                ctypes.byref(written), None))
            CHK(nidaq.DAQmxStartTask(taskHandle))
    except Exception, e:
        logging.error('NI DAQ call failed (correct channel configuration selected?): %s', str(e))
    finally:
        if taskHandle.value != 0:
            nidaq.DAQmxStopTask(taskHandle)
            nidaq.DAQmxClearTask(taskHandle)

    return written.value

def read_counter(devchan="/Dev1/ctr0", samples=1, freq=1.0, timeout=1.0, src=""):
    '''
    Read counter 'devchan'.
    Specify source pin with 'src'.
    '''

    taskHandle = TaskHandle(0)
    try:
        CHK(nidaq.DAQmxCreateTask("", ctypes.byref(taskHandle)))
        initial_count = int32(0)
        CHK(nidaq.DAQmxCreateCICountEdgesChan(taskHandle, devchan, "",
                DAQmx_Val_Rising, initial_count, DAQmx_Val_CountUp))
        if src is not None and src != "":
            CHK(nidaq.DAQmxSetCICountEdgesTerm(taskHandle, devchan, src))

        nread = int32()
        data = numpy.zeros(samples, dtype=numpy.float64)
        if samples > 1:
            CHK(nidaq.DAQmxCfgSampClkTiming(taskHandle, "", float64(freq),
                DAQmx_Val_Rising, DAQmx_Val_FiniteSamps,
                uInt64(samples)));
            CHK(nidaq.DAQmxStartTask(taskHandle))
            CHK(nidaq.DAQmxReadAnalogF64(taskHandle, int32(samples), float64(timeout),
               DAQmx_Val_GroupByChannel, data.ctypes.data,
               samples, ctypes.byref(read), None))
        else:
            CHK(nidaq.DAQmxStartTask(taskHandle))
            time.sleep(1.0 / freq)
            nread = int32(0)
            CHK(nidaq.DAQmxReadCounterF64(taskHandle, int32(samples), float64(timeout),
                data.ctypes.data, int32(samples), ctypes.byref(nread), None))
            nread = int32(1)

    except Exception, e:
        logging.error('NI DAQ call failed: %s', str(e))

    finally:
        if taskHandle.value != 0:
            nidaq.DAQmxStopTask(taskHandle)
            nidaq.DAQmxClearTask(taskHandle)

    if nread.value == 1:
        return int(data[0])
    else:
        return data

def create_counter_task(devchan, samples=1, freq=1, timeout=1, src=""):
    taskHandle = TaskHandle(0)

    try:
        CHK(nidaq.DAQmxCreateTask("", ctypes.byref(taskHandle)))
        initial_count = int32(0)
        CHK(nidaq.DAQmxCreateCICountEdgesChan(taskHandle, devchan, "",
                DAQmx_Val_Rising, initial_count, DAQmx_Val_CountUp))
        if src is not None and src != "":
            CHK(nidaq.DAQmxSetCICountEdgesTerm(taskHandle, devchan, src))

        if samples > 1:
            CHK(nidaq.DAQmxCfgSampClkTiming(taskHandle, "", float64(freq),
                DAQmx_Val_Rising, DAQmx_Val_FiniteSamps,
                uInt64(samples)));

    except Exception, e:
        logging.error('NI DAQ call failed: %s', str(e))
        if taskHandle.value != 0:
            nidaq.DAQmxStopTask(taskHandle)
            nidaq.DAQmxClearTask(taskHandle)

    return taskHandle

def read_counters(devchans=["/Dev1/ctr0","/Dev1/ctr1"], samples=1, freq=1.0, timeout=1.0, src=None):
    tasks = []
    devsrc = None
    ret = []
    for i, dev in enumerate(devchans):
        if src is not None:
            devsrc = src[i]
        result = create_counter_task(dev, samples, freq, timeout, devsrc)
        if result != -1:
            tasks.append(result)

    try:
        for task in tasks:
            CHK(nidaq.DAQmxStartTask(task))

        time.sleep(float(samples) / freq)

        for task in tasks:
            data = numpy.zeros(samples, dtype=numpy.float64)
            if samples > 1:
                CHK(nidaq.DAQmxReadAnalogF64(task, int32(samples), float64(timeout),
                        DAQmx_Val_GroupByChannel, data.ctypes.data,
                        samples, ctypes.byref(read), None))
                ret.append(data)
            else:
                nread = int32(0)
                CHK(nidaq.DAQmxReadCounterF64(task, int32(samples), float64(timeout),
                    data.ctypes.data, int32(samples), ctypes.byref(nread), None))
                nread = int32(1)
                ret.append(data[0])

    except Exception, e:
        logging.error('NI DAQ call failed: %s', str(e))

    finally:
        for task in tasks:
            if task.value != 0:
                nidaq.DAQmxStopTask(task)
                nidaq.DAQmxClearTask(task)

    return ret

def write_dig_port8(channel, val, timeout=1.0):
    '''
    Set digital output of channels.
    The value is sent to the specified channels, LSB to MSB.
    '''

    taskHandle = TaskHandle(0)
    try:
        CHK(nidaq.DAQmxCreateTask("", ctypes.byref(taskHandle)))
        CHK(nidaq.DAQmxCreateDOChan(taskHandle, channel, '', DAQmx_Val_ChanForAllLines))

        nwritten = int32(0)

#        val = numpy.array((val,), dtype=numpy.int16)
#       This requires shifting bits if writing to part of a port
#        CHK(nidaq.DAQmxWriteDigitalU16(taskHandle, int32(1), int32(1),
#            float64(1.0), int32(DAQmx_Val_GroupByChannel), val.ctypes.data, ctypes.byref(nwritten), None))

        vals = numpy.array([(val >> i) & 1 for i in range(8)], dtype=numpy.int8)
        nbytes = int32(0)
        CHK(nidaq.DAQmxGetWriteDigitalLinesBytesPerChan(taskHandle, ctypes.byref(nbytes)))
        CHK(nidaq.DAQmxWriteDigitalLines(taskHandle, int32(1), int32(1),
            float64(1.0), int32(DAQmx_Val_GroupByChannel), vals.ctypes.data, ctypes.byref(nwritten), None))

        CHK(nidaq.DAQmxStartTask(taskHandle))

    except Exception, e:
        logging.error('NI DAQ call failed: %s', str(e))

    finally:
        if taskHandle.value != 0:
            nidaq.DAQmxStopTask(taskHandle)
            nidaq.DAQmxClearTask(taskHandle)


def write_dig_port_stream(channel, samples, rate, timeout=1.0):
    '''
    Output a binary waveform
    
    Input:
        channel - channels to output data to
        samples - a 2d numpy array of booleans, 2nd index is the channel
        rate - sample rate in Hz
    '''

    taskHandle = TaskHandle(0)
    try:
        CHK(nidaq.DAQmxCreateTask("", ctypes.byref(taskHandle)))
        CHK(nidaq.DAQmxCreateDOChan(taskHandle, channel, '', DAQmx_Val_ChanForAllLines))

        nwritten = int32(0)

        vals = numpy.array(samples, numpy.bool8)
        if (vals.ndim == 1):
            nSamples = 1
        else:
            nSamples = vals.shape[0]
        # DAQmxSetSampClkSrc(taskHandle, "")
        CHK(nidaq.DAQmxCfgSampClkTiming(taskHandle, "", float64(rate), DAQmx_Val_Rising, DAQmx_Val_FiniteSamps, uInt64(nSamples)))
        nbytes = int32(0)
        CHK(nidaq.DAQmxGetWriteDigitalLinesBytesPerChan(taskHandle, ctypes.byref(nbytes)))
        CHK(nidaq.DAQmxWriteDigitalLines(taskHandle, int32(nSamples), int32(1),
            float64(1.0), int32(DAQmx_Val_GroupByChannel), vals.ctypes.data, ctypes.byref(nwritten), None))
        CHK(nidaq.DAQmxStartTask(taskHandle))

    except Exception, e:
        logging.error('NI DAQ call failed: %s', str(e))

    finally:
        if taskHandle.value != 0:
            nidaq.DAQmxStopTask(taskHandle)
            nidaq.DAQmxClearTask(taskHandle)



# qtLAB driver for microwave sources Anritsu MG37022, Anritsu MG3692C (*)
# filename: Anritsu_MG37022.py
# Pascal Macha <pascalmacha@googlemail.com>, 2010
# Jochen Braumueller <jochen.braumueller@kit.edu> 2016

# (*) phase offset functions not supported by Anritsu MG3692C

#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import qkit
from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
import math
from time import sleep

class Anritsu_MG37022(Instrument):
    '''
    This is the python driver for the Anritsu_MG37022 microwave source

    Usage:
    Initialise with
    <name> = instruments.create('<name>', address='<GPIB address>')
    '''

    def __init__(self, name, address, model = 'Anritsu'):
        '''
        Initializes the Anritsu_MG37022, and communicates with the wrapper.
        
        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
        '''
        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._model = model
        self._visainstrument = visa.instrument(self._address)
        self._slope = None   #set _slope to True if frequency dependent power compensation is requested
        sleep(1)

        # Implement parameters
        self.add_parameter('frequency', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz',tags=['sweep'])
        self.add_parameter('phase_offset', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=-360, maxval=360,
            units='deg')
        self.add_parameter('slope', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=100,
            units='dB@10GHz')
        self.add_parameter('power', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=-130, maxval=29,   #29dBm possible, this is security JB
            units='dBm', tags=['sweep'])
        self.add_parameter('status', type=bool,
            flags=Instrument.FLAG_GETSET)
        self.add_parameter('high_power', type=bool,
            flags=Instrument.FLAG_GETSET)
        
        # Implement functions
        self.add_function('get_all')
        self.get_all()
        
    # initialization related
    def get_all(self):
        self.get_frequency()
        self.get_power()
        self.get_status()

    #Communication with device
    def do_get_frequency(self):
        '''
        Get frequency of device

        Input:
            -

        Output:
            microwave frequency (Hz)
        '''
        self._frequency = float(self.ask('SOUR:FREQ:CW?'))
        return self._frequency
        
    def do_set_frequency(self, frequency):
        '''
        Set frequency of device

        Input:
            freq (float) : Frequency in Hz

        Output:
            None
        '''
        #logging.debug(__name__ + ' : setting frequency to %s Hz' % (frequency*1.0e9))
        self.write('SOUR:FREQ:CW %i' % (int(frequency)))
        self._frequency = float(frequency)
        if(self._slope != None): 
            self.do_set_power()
            
    def do_get_slope(self):
        '''
        Get attribute '_slope'

        Input:
            -

        Output:
            slope applied to calculate power values
            None if no slope is set
        '''
        return self._slope
        
    def do_set_slope(self, slope):
        '''
        Set slope of output power to use at different frequencies.
        Assumes a cable attenuation of attn[dB] ~ attn(f0)*sqrt(f/f0), f0=10GHz

        Input:
            slope (float) : Slope to apply to power values.
        '''
        self._slope = slope
        
    def do_get_power(self):
        '''
        Get output power of device

        Input:
            -

        Output:
            microwave power (dBm)
        '''
        self._power = float(self.ask('POW:LEV?'))
        return self._power
        
    def do_set_power(self,power = None):
        '''
        Set output power of device, limited to +12dBm
        upper bound can be increased by setting self.enable_high_power to True

        Input:
            power (float) : Power in dBm
            If a slope is set, this is the power at 10GHz.

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting power to %s dBm' % power)
        if(power == None): #This block is needed for the slope function.
            power = self._power
        else:
            self._power = power
        # compensate cables
        if(self._slope != None):
            power = power+self._slope*(math.sqrt(self._frequency/10.e9)-1)
        self.write('POW:LEV %.2f' % power)
            
    def do_get_status(self):
        '''
        Get status of output channel

        Input:
            -

        Output:
            True (on) or False (off)
        '''
        stat = bool(int(self.ask('OUTP:STAT?')))
        return stat
        
    def do_set_status(self,status):
        '''
        Set status of output channel

        Input:
            status : True or False

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting status to "%s"' % status)
        if status == True:
            self.write('OUTP:STAT ON')
        elif status == False:
            self.write('OUTP:STAT OFF')
        else:
            logging.error('set_status(): can only set True or False')
        
        
    def do_get_phase_offset(self):
        '''
        Get the RF output phase offset

        Input:
            -

        Output:
            phase offset (deg)
        '''
        return float(self.ask('PHAS:ADJ?'))
        
    def do_set_phase_offset(self, phase_offset):
        '''
        Set the RF output phase offset and display phase offset

        Input:
            phase offset in degrees (-360.0 .. 360.0)

        Output:
            Nones
        '''
        self.write('PHAS:ADJ %.1f DEG' %phase_offset)
        self.write('PHAS:DISP ON')
            
    def do_set_high_power(self, enabled):
        '''
        As default, the power parameter is limited to +12dB, to avoid
        any damage by careless use.
        You can enable the high power mode to increase the power to +30dBm.
        Be careful.
        
        Input:
            enabled: True or False
        '''
        if enabled:
            self.set_parameter_bounds('power',-20,30)
        else:
            self.set_parameter_bounds('power',-130,12)
            
    def do_get_high_power(self):
        '''
        Gets the current state of the high_power option.
        '''
        return self.get_parameter_options('power')['maxval'] > 13
            
    #shortcuts
    def off(self):
        '''
        Set status to 'off'

        Input:
            None

        Output:
            None
        '''
        self.set_status(False)

    def on(self):
        '''
        Set status to 'on'

        Input:
            None

        Output:
            None
        '''
        self.set_status(True)
        
    #sending customized messages
    def write(self,msg):
      return self._visainstrument.write(msg)
    
    if qkit.visa.qkit_visa_version == 1:
        def ask(self,msg):
            return self._visainstrument.ask(msg)
    else:
        def ask(self, msg):
            return self._visainstrument.query(msg)



from abc import ABC, abstractmethod
from qkit.core.instrument_base import Instrument
import numpy as np

class AbstractVNA(ABC):
    """
    This Abstract Base Class defines all the methods required and optionally available for VNA measurements.
    This Abstract Base Class expects the object to be a Instrument, but will not directly enforce this.
    We check the presence of the required methods before calling them.
    """

    def register_vna_functions(self):
        assert isinstance(self, Instrument)
        self.add_function('get_freqpoints')
        self.add_function('get_tracedata')
        self.add_function('get_sweeptime')
        self.add_function('get_sweeptime_averages')
        self.add_function('pre_measurement')
        self.add_function('start_measurement')
        self.add_function('ready')
        self.add_function('post_measurement')

    @abstractmethod
    def get_freqpoints(self) -> np.ndarray:
        """
        Return a list of frequencies the VNA is measuring at.
        This may be configurable with device specific methods.
        """
        pass

    def get_span(self) -> float:
        """
        Returns the length of the frequency interval to be analyzed.
        """
        freqs = self.get_freqpoints()
        return freqs[-1] - freqs[0]

    @abstractmethod
    def get_tracedata(self, RealImag = None) -> tuple((np.ndarray, np.ndarray)):
        """
        When the measurement succeeded, this method returns the resulting data.
        The behaviour depends on the [RealImag] keyword argument.
        If it is none, this method shall return (Amplitude, Phase)-Data.
        Otherwise, this method must return (I, Q)-Data.
        """
        pass

    @abstractmethod
    def get_sweeptime(self, query=True):
        """
        Report the expected sweep time for the GUI.
        """
        pass

    @abstractmethod
    def get_sweeptime_averages(self):
        """
        No idead what this does. TODO: Figure out.
        """
        pass

    
    def pre_measurement(self):
        """
        Optional hook to execute code before a measurement.
        If the device needs to be set up, between measurements,
        this is the point where you can do that.
        """
        pass

    @abstractmethod
    def start_measurement(self):
        """
        Trigger the actual measurement. If this is synchronous [ready] may return True.
        Otherwise, use [ready] to indicate when the measurement has completed.
        """
        pass

    @abstractmethod
    def ready(self) -> bool:
        """
        Report if this measurement has completed. If start_measurement returns on completion,
        this should be implmented to always return true.
        """
        pass

    def post_measurement(self):
        """
        Optional hook to execute code after a measurement has completed.
        """
        pass

    def get_all(self):
        """
        Support to efficiently retrieve all options about this device.
        """
        pass

    def get_nop(self):
        return len(self.get_freqpoints())

    def get_averages(self):
        return 1

    def get_Average(self):
        return False

    def avg_clear(self):
        pass